import itertools
from typing import TYPE_CHECKING

import numpy
import pytest
from openff.toolkit import ForceField, Molecule, Quantity, Topology
from openff.toolkit.typing.engines.smirnoff.parameters import (
    ElectrostaticsHandler,
    LibraryChargeHandler,
    VirtualSiteHandler,
    vdWHandler,
)
from openff.units import Unit
from openff.units.openmm import to_openmm
from openff.utilities import has_package, skip_if_missing

if has_package("openmm") or TYPE_CHECKING:
    import openmm
    import openmm.unit


def _get_interpolated_bond_k(bond_handler) -> float:
    for key in bond_handler.key_map:
        if key.bond_order is not None:
            topology_key = key
            break
    potential_key = bond_handler.key_map[topology_key]
    return bond_handler.potentials[potential_key].parameters["k"].m


class TestSMIRNOFFVirtualSiteCharges:
    @pytest.mark.parametrize("chlorine_charge", [-0.1, 0.22, 1.3])
    def test_neutral_total_charge(self, sage_with_bond_charge, chlorine_charge):
        sage_with_bond_charge.deregister_parameter_handler("NAGLCharges")
        sage_with_bond_charge.get_parameter_handler("ChargeIncrementModel")
        sage_with_bond_charge["ChargeIncrementModel"].partial_charge_method = "zeros"

        sage_with_bond_charge["VirtualSites"].get_parameter(
            parameter_attrs={
                "smirks": "[#6:2]-[#17X1:1]",
            },
        )[0].charge_increment1 = Quantity(
            chlorine_charge,
            "elementary_charge",
        )

        out = sage_with_bond_charge.create_interchange(
            Molecule.from_mapped_smiles(
                "[H:3][C:1]([H:4])([H:5])[Cl:2]",
            ).to_topology(),
        )

        assert {key.virtual_site_type for key in out["VirtualSites"].potentials} == {
            "BondCharge",
        }

        charges = [charge.m for charge in out["Electrostatics"]._get_charges().values()]

        assert sum(charges) == 0.0

        # The carbon's charge was not modified
        assert charges[0] == 0.0

        # The chlorine received variable charge from the virtual site, which should have the negative of it
        # SMIRNOFF spec (Aug 2023):
        #   Each virtual site receives charge which is transferred from the desired atoms specified
        #   in the SMIRKS pattern via a charge_increment# parameter,
        #   e.g., if charge_increment1=0.1*elementary_charge then the virtual site will receive a
        #   charge of -0.1 and the atom labeled 1 will have its charge adjusted upwards by 0.1.
        assert charges[1] == chlorine_charge
        assert charges[-1] == -1 * chlorine_charge

        charges_without_virtual_sites = [
            charge.m for charge in out["Electrostatics"]._charges_without_virtual_sites.values()
        ]

        assert sum(charges_without_virtual_sites) == chlorine_charge != 0.0


@skip_if_missing("openmm")
class TestSMIRNOFFVirtualSites:
    from openff.toolkit._tests.mocking import VirtualSiteMocking

    @classmethod
    def build_force_field(cls, v_site_handler: VirtualSiteHandler) -> ForceField:
        force_field = ForceField()

        force_field.get_parameter_handler("Bonds").add_parameter(
            {
                "smirks": "[*:1]~[*:2]",
                "k": Quantity(0.0, "kilojoule_per_mole / angstrom**2"),
                "length": Quantity(0.0, "angstrom"),
            },
        )
        force_field.get_parameter_handler("Angles").add_parameter(
            {
                "smirks": "[*:1]~[*:2]~[*:3]",
                "k": Quantity(0.0, "kilojoule_per_mole / degree**2"),
                "angle": Quantity(60.0, "degree"),
            },
        )
        force_field.get_parameter_handler("ProperTorsions").add_parameter(
            {
                "smirks": "[*:1]~[*:2]~[*:3]~[*:4]",
                "k": Quantity([0.0], "kilojoule_per_mole"),
                "phase": Quantity([0.0], "degree"),
                "periodicity": [2],
                "idivf": [1.0],
            },
        )
        force_field.get_parameter_handler("vdW").add_parameter(
            {
                "smirks": "[*:1]",
                "epsilon": Quantity(0.0, "kilojoule_per_mole"),
                "sigma": Quantity(1.0, "angstrom"),
            },
        )
        force_field.get_parameter_handler("LibraryCharges").add_parameter(
            {"smirks": "[*:1]", "charge": Quantity([0.0], "elementary_charge")},
        )
        force_field.get_parameter_handler("Electrostatics")

        force_field.register_parameter_handler(v_site_handler)

        return force_field

    @classmethod
    def generate_v_site_coordinates(
        cls,
        molecule: Molecule,
        input_conformer: Quantity,
        parameter: VirtualSiteHandler.VirtualSiteType,
    ) -> Quantity:
        # Compute the coordinates of the virtual site. Unfortunately OpenMM does not
        # seem to offer a more compact way to do this currently.
        handler = VirtualSiteHandler(version="0.3")
        handler.add_parameter(parameter=parameter)

        force_field = cls.build_force_field(handler)

        system = force_field.create_openmm_system(molecule.to_topology())

        n_v_sites = sum(1 if system.isVirtualSite(i) else 0 for i in range(system.getNumParticles()))

        input_conformer = Quantity(
            numpy.vstack(
                [
                    input_conformer.m_as("angstrom"),
                    numpy.zeros((n_v_sites, 3)),
                ],
            ),
            "angstrom",
        )

        context = openmm.Context(
            system,
            openmm.LangevinMiddleIntegrator(
                300 * openmm.unit.kelvin,
                1 / openmm.unit.picosecond,
                2.0 * openmm.unit.femtosecond,
            ),
            openmm.Platform.getPlatformByName("Reference"),
        )

        context.setPositions(to_openmm(input_conformer))
        context.computeVirtualSites()

        output_conformer = context.getState(getPositions=True).getPositions(
            asNumpy=True,
        )

        return output_conformer[molecule.n_atoms :, :]

    @pytest.mark.skip(
        "Interchange does not allow this combination of non-bonded settings.",
    )
    @pytest.mark.parametrize(
        (
            "parameter",
            "smiles",
            "input_conformer",
            "atoms_to_shuffle",
            "expected_coordinates",
        ),
        [
            (
                VirtualSiteMocking.bond_charge_parameter("[Cl:1]-[C:2]"),
                "[Cl:1][C:2]([H:3])([H:4])[H:5]",
                VirtualSiteMocking.sp3_conformer(),
                (0, 1),
                Quantity(numpy.array([[0.0, 3.0, 0.0]]), "angstrom"),
            ),
            (
                VirtualSiteMocking.bond_charge_parameter("[C:1]#[C:2]"),
                "[H:1][C:2]#[C:3][H:4]",
                VirtualSiteMocking.sp1_conformer(),
                (2, 3),
                Quantity(
                    numpy.array([[-3.0, 0.0, 0.0], [3.0, 0.0, 0.0]]),
                    "angstrom",
                ),
            ),
            (
                VirtualSiteMocking.monovalent_parameter("[O:1]=[C:2]-[H:3]"),
                "[O:1]=[C:2]([H:3])[H:4]",
                VirtualSiteMocking.sp2_conformer(),
                (0, 1, 2, 3),
                (
                    VirtualSiteMocking.sp2_conformer()[0]
                    + Quantity(
                        numpy.array(
                            [[1.0, numpy.sqrt(2), 1.0], [1.0, -numpy.sqrt(2), -1.0]],
                        ),
                        "angstrom",
                    )
                ),
            ),
            (
                VirtualSiteMocking.divalent_parameter(
                    "[H:2][O:1][H:3]",
                    match="once",
                    angle=Quantity(0.0, "degree"),
                ),
                "[H:2][O:1][H:3]",
                VirtualSiteMocking.sp2_conformer()[1:, :],
                (0, 1, 2),
                Quantity(numpy.array([[2.0, 0.0, 0.0]]), "angstrom"),
            ),
            (
                VirtualSiteMocking.divalent_parameter(
                    "[H:2][O:1][H:3]",
                    match="all_permutations",
                    angle=Quantity(45.0, "degree"),
                ),
                "[H:2][O:1][H:3]",
                VirtualSiteMocking.sp2_conformer()[1:, :],
                (0, 1, 2),
                Quantity(
                    numpy.array(
                        [
                            [numpy.sqrt(2), numpy.sqrt(2), 0.0],
                            [numpy.sqrt(2), -numpy.sqrt(2), 0.0],
                        ],
                    ),
                    "angstrom",
                ),
            ),
            (
                VirtualSiteMocking.trivalent_parameter("[N:1]([H:2])([H:3])[H:4]"),
                "[N:1]([H:2])([H:3])[H:4]",
                VirtualSiteMocking.sp3_conformer()[1:, :],
                (0, 1, 2, 3),
                Quantity(numpy.array([[0.0, 2.0, 0.0]]), "angstrom"),
            ),
        ],
    )
    def test_v_site_geometry(
        self,
        parameter: VirtualSiteHandler.VirtualSiteType,
        smiles: str,
        input_conformer: Quantity,
        atoms_to_shuffle: tuple[int, ...],
        expected_coordinates: Quantity,
    ):
        """An integration test that virtual sites are placed correctly relative to the
        parent atoms"""

        for atom_permutation in itertools.permutations(atoms_to_shuffle):
            molecule = Molecule.from_mapped_smiles(smiles, allow_undefined_stereo=True)
            molecule._conformers = [input_conformer]

            # We shuffle the ordering of the atoms involved in the v-site orientation
            # to ensure that the orientation remains invariant as expected.
            shuffled_atom_order = {i: i for i in range(molecule.n_atoms)}
            shuffled_atom_order.update(
                {old_index: new_index for old_index, new_index in zip(atoms_to_shuffle, atom_permutation)},
            )

            molecule = molecule.remap(shuffled_atom_order)

            output_coordinates = self.generate_v_site_coordinates(
                molecule,
                molecule.conformers[0],
                parameter,
            )

            assert output_coordinates.shape == expected_coordinates.shape

            def sort_coordinates(x: numpy.ndarray) -> numpy.ndarray:
                # Sort the rows by first, then second, then third columns as row
                # as the order of v-sites is not deterministic.
                return x[numpy.lexsort((x[:, 2], x[:, 1], x[:, 0])), :]

            found = sort_coordinates(
                output_coordinates.value_in_unit(openmm.unit.angstrom),
            )
            expected = sort_coordinates(expected_coordinates.m_as("angstrom"))

            assert numpy.allclose(found, expected), expected - found

    _E = Unit("elementary_charge")
    _A = Unit("angstrom")
    _KJ = Unit("kilojoule_per_mole")

    @pytest.mark.skip(
        "Interchange does not allow this combination of non-bonded settings.",
    )
    @pytest.mark.parametrize(
        ("topology", "parameters", "expected_parameters", "expected_n_v_sites"),
        [
            (
                Topology.from_molecules(
                    [
                        VirtualSiteMocking.chloromethane(reverse=False),
                        VirtualSiteMocking.chloromethane(reverse=True),
                    ],
                ),
                [VirtualSiteMocking.bond_charge_parameter("[Cl:1][C:2]")],
                [
                    # charge, sigma, epsilon
                    (0.1 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.2 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.0 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.0 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.0 * _E, 10.0 * _A, 0.0 * _KJ),
                    (-0.3 * _E, 4.0 * _A, 3.0 * _KJ),
                    (0.0 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.0 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.0 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.2 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.1 * _E, 10.0 * _A, 0.0 * _KJ),
                    (-0.3 * _E, 4.0 * _A, 3.0 * _KJ),
                ],
                2,
            ),
            # Check a complex case of bond charge vsite assignment and overwriting
            (
                Topology.from_molecules(
                    [
                        VirtualSiteMocking.formaldehyde(reverse=False),
                        VirtualSiteMocking.formaldehyde(reverse=True),
                    ],
                ),
                [
                    VirtualSiteMocking.bond_charge_parameter("[O:1]=[*:2]"),
                    VirtualSiteMocking.bond_charge_parameter(
                        "[O:1]=[C:2]",
                        param_multiple=1.5,
                    ),
                    VirtualSiteMocking.bond_charge_parameter(
                        "[O:1]=[CX3:2]",
                        param_multiple=2.0,
                        name="EP2",
                    ),
                ],
                [
                    # charge, sigma, epsilon
                    (0.35 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.7 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.0 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.0 * _E, 10.0 * _A, 0.0 * _KJ),
                    (-0.45 * _E, 6.0 * _A, 4.5 * _KJ),  # C=O vsite
                    (-0.6 * _E, 8.0 * _A, 6.0 * _KJ),  # CX3=O vsite
                    (0.0 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.0 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.7 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.35 * _E, 10.0 * _A, 0.0 * _KJ),
                    (-0.45 * _E, 6.0 * _A, 4.5 * _KJ),  # C=O vsite
                    (-0.6 * _E, 8.0 * _A, 6.0 * _KJ),  # CX3=O vsite
                ],
                4,
            ),
            (
                Topology.from_molecules(
                    [
                        VirtualSiteMocking.formaldehyde(reverse=False),
                        VirtualSiteMocking.formaldehyde(reverse=True),
                    ],
                ),
                [VirtualSiteMocking.monovalent_parameter("[O:1]=[C:2]-[H:3]")],
                [
                    # charge, sigma, epsilon
                    (0.2 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.4 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.3 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.3 * _E, 10.0 * _A, 0.0 * _KJ),
                    (-0.6 * _E, 4.0 * _A, 5.0 * _KJ),
                    (-0.6 * _E, 4.0 * _A, 5.0 * _KJ),
                    (0.3 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.3 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.4 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.2 * _E, 10.0 * _A, 0.0 * _KJ),
                    (-0.6 * _E, 4.0 * _A, 5.0 * _KJ),
                    (-0.6 * _E, 4.0 * _A, 5.0 * _KJ),
                ],
                4,
            ),
            (
                Topology.from_molecules(
                    [
                        VirtualSiteMocking.hypochlorous_acid(reverse=False),
                        VirtualSiteMocking.hypochlorous_acid(reverse=True),
                    ],
                ),
                [
                    VirtualSiteMocking.divalent_parameter(
                        "[H:2][O:1][Cl:3]",
                        match="all_permutations",
                    ),
                ],
                [
                    # charge, sigma, epsilon
                    (0.2 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.1 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.3 * _E, 10.0 * _A, 0.0 * _KJ),
                    (-0.6 * _E, 4.0 * _A, 5.0 * _KJ),
                    (0.3 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.1 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.2 * _E, 10.0 * _A, 0.0 * _KJ),
                    (-0.6 * _E, 4.0 * _A, 5.0 * _KJ),
                ],
                2,
            ),
            (
                Topology.from_molecules(
                    [
                        VirtualSiteMocking.fake_ammonia(reverse=False),
                        VirtualSiteMocking.fake_ammonia(reverse=True),
                    ],
                ),
                [VirtualSiteMocking.trivalent_parameter("[N:1]([Br:2])([Cl:3])[H:4]")],
                [
                    # charge, sigma, epsilon
                    (0.1 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.3 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.2 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.4 * _E, 10.0 * _A, 0.0 * _KJ),
                    (-1.0 * _E, 5.0 * _A, 6.0 * _KJ),
                    (0.4 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.2 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.3 * _E, 10.0 * _A, 0.0 * _KJ),
                    (0.1 * _E, 10.0 * _A, 0.0 * _KJ),
                    (-1.0 * _E, 5.0 * _A, 6.0 * _KJ),
                ],
                2,
            ),
        ],
    )
    def test_create_force(
        self,
        topology: Topology,
        parameters: list[VirtualSiteHandler.VirtualSiteType],
        expected_parameters: list[
            tuple[
                "openmm.unit.Quantity",
                "openmm.unit.Quantity",
                "openmm.unit.Quantity",
            ]
        ],
        expected_n_v_sites: int,
    ):
        expected_n_total = topology.n_atoms + expected_n_v_sites
        # sanity check the test input
        assert len(expected_parameters) == expected_n_total

        handler = VirtualSiteHandler(version="0.3")

        for parameter in parameters:
            handler.add_parameter(parameter=parameter)

        force_field = ForceField()

        force_field.register_parameter_handler(ElectrostaticsHandler(version=0.4))
        force_field.register_parameter_handler(LibraryChargeHandler(version=0.3))
        force_field.get_parameter_handler("LibraryCharges").add_parameter(
            {
                "smirks": "[*:1]",
                "charge": Quantity([0.0], "elementary_charge"),
            },
        )
        force_field.register_parameter_handler(vdWHandler(version=0.3))
        force_field.get_parameter_handler("vdW").add_parameter(
            {
                "smirks": "[*:1]",
                "epsilon": Quantity(0.0, "kilojoule_per_mole"),
                "sigma": Quantity(1.0, "nanometer"),
            },
        )
        force_field.register_parameter_handler(handler)

        assert {
            key.virtual_site_type for key in force_field.create_interchange(topology)["VirtualSites"].potentials
        } == {parameter.type for parameter in parameters}

        system = force_field.create_openmm_system(topology)

        assert system.getNumParticles() == expected_n_total

        # TODO: Explicitly ensure virtual sites are collated between moleucles
        #       This is implicitly tested in the construction of the parameter arrays

        assert system.getNumForces() == 1
        force: openmm.NonbondedForce = next(iter(system.getForces()))

        total_charge = 0.0 * openmm.unit.elementary_charge

        for i, (expected_charge, expected_sigma, expected_epsilon) in enumerate(
            expected_parameters,
        ):
            charge, sigma, epsilon = force.getParticleParameters(i)

            # Make sure v-sites are massless.
            assert (numpy.isclose(system.getParticleMass(i)._value, 0.0)) == system.isVirtualSite(i)

            assert numpy.isclose(
                expected_charge.m_as("elementary_charge"),
                charge.value_in_unit(openmm.unit.elementary_charge),
            )
            assert numpy.isclose(
                expected_sigma.m_as("angstrom"),
                sigma.value_in_unit(openmm.unit.angstrom),
            )
            assert numpy.isclose(
                expected_epsilon.m_as("kilojoule_per_mole"),
                epsilon.value_in_unit(openmm.unit.kilojoule_per_mole),
            )

            total_charge += charge

        expected_total_charge = sum(molecule.total_charge.m_as("elementary_charge") for molecule in topology.molecules)

        assert numpy.isclose(
            expected_total_charge,
            total_charge.value_in_unit(openmm.unit.elementary_charge),
        )

    def test_virtual_site_type_stored_in_potential_key(
        self,
        ethanol,
        sage_with_bond_charge,
        sage_with_trivalent_nitrogen,
    ):
        # Can't use a fixture here because of the modified versions
        sage = ForceField("openff-2.1.0.offxml")

        assert {key.virtual_site_type for key in sage.create_interchange(ethanol.to_topology())["vdW"].potentials} == {
            None,
        }

        with pytest.raises(LookupError):
            assert {
                key.virtual_site_type
                for key in sage.create_interchange(ethanol.to_topology())["VirtualSites"].potentials
            } == {None}

        assert {
            key.virtual_site_type
            for key in sage_with_bond_charge.create_interchange(
                Molecule.from_smiles("CCl").to_topology(),
            )["VirtualSites"].potentials
        } == {"BondCharge"}

        assert {
            key.virtual_site_type
            for key in sage_with_trivalent_nitrogen.create_interchange(
                Molecule.from_smiles("N").to_topology(),
            )["VirtualSites"].potentials
        } == {"TrivalentLonePair"}

    def test_identical_smirks_do_not_clash(
        self,
        sage_with_two_virtual_sites_same_smirks,
    ):
        """Reproduce issue #940."""
        nitrogen = Molecule.from_mapped_smiles("[H:2][N:1]([H:3])[H:4]")

        out = sage_with_two_virtual_sites_same_smirks.create_interchange(
            nitrogen.to_topology(),
        )

        charges = [charge.m for charge in out["Electrostatics"]._get_charges().values()]

        assert sum(charges) == pytest.approx(0.0)

        # The second parameter shifts 5 e of charge from the nitrogen to the
        # virtual site, so the nitrogen should end up with a VERY positive charge. If
        # the parameter collision in the issue happens, this should closer to +1.0.
        # Use rough numbers because of AM1-BCC inconsistency.
        assert charges[0] > 4.0

        # charges on virtual sites should not match
        assert charges[-2] != charges[-1]



from copy import deepcopy

import numpy
import pytest
from openff.toolkit import Molecule, Quantity, Topology
from pydantic import ValidationError

from openff.interchange import Interchange
from openff.interchange._tests import needs_gmx
from openff.interchange.components.mdconfig import get_intermol_defaults
from openff.interchange.drivers.gromacs import _process, _run_gmx_energy
from openff.interchange.interop.gromacs.models.models import GROMACSAtomType
from openff.interchange.smirnoff._gromacs import _convert


@pytest.fixture
def molecule1():
    molecule = Molecule.from_smiles(
        "[H][O][c]1[c]([H])[c]([O][H])[c]([H])[c]([O][H])[c]1[H]",
    )
    molecule.generate_conformers(n_conformers=1)
    molecule.name = "MOL__1"

    return molecule


@pytest.fixture
def molecule2():
    molecule = Molecule.from_smiles("C1=C(C=C(C=C1C(=O)O)C(=O)O)C(=O)O")
    molecule.generate_conformers(n_conformers=1)
    molecule.name = "MOL__2"

    molecule.conformers[0] += Quantity(numpy.array([5, 0, 0]), "angstrom")

    return molecule


@pytest.fixture
def system1(molecule1, sage):
    box = Quantity(5 * numpy.eye(3), "nanometer")

    return _convert(Interchange.from_smirnoff(sage, [molecule1], box=box))


@pytest.fixture
def system2(molecule2, sage):
    box = Quantity(5 * numpy.eye(3), "nanometer")

    return _convert(Interchange.from_smirnoff(sage, [molecule2], box=box))


@pytest.fixture
def combined_system(molecule1, molecule2, sage):
    box = Quantity(5 * numpy.eye(3), "nanometer")

    return _convert(
        Interchange.from_smirnoff(
            sage,
            Topology.from_molecules([molecule1, molecule2]),
            box=box,
        ),
    )


class TestModels:
    def test_massless_atom_error(self):
        with pytest.raises(
            ValidationError,
            match=r"Particle type.*D.*V.*if massless",
        ):
            GROMACSAtomType(
                name="foo",
                bonding_type="",
                atomic_number=100,
                mass=Quantity(0.0, "dalton"),
                charge=Quantity(0.0, "elementary_charge"),
                particle_type="A",
            )

    def test_massive_virtual_site_error(self):
        with pytest.raises(
            ValidationError,
            match=r"Particle type.*A.*if it has mass",
        ):
            GROMACSAtomType(
                name="foo",
                bonding_type="",
                atomic_number=100,
                mass=Quantity(20.0, "dalton"),
                charge=Quantity(0.0, "elementary_charge"),
                particle_type="V",
            )


class TestAddRemoveMoleculeType:
    @needs_gmx
    @pytest.mark.parametrize("molecule_name", ["MOL__1", "MOL__2"])
    def test_remove_basic(self, combined_system, molecule_name):
        combined_system.remove_molecule_type(molecule_name)

        # Just a sanity check
        combined_system.to_files(prefix=molecule_name, decimal=8)

        get_intermol_defaults(periodic=True).write_mdp_file("tmp.mdp")

        _process(
            _run_gmx_energy(f"{molecule_name}.top", f"{molecule_name}.gro", "tmp.mdp"),
        )

    @pytest.mark.slow
    @pytest.mark.parametrize("molecule_name", ["MOL__1", "MOL__2"])
    def test_add_existing_molecule_type(self, combined_system, molecule_name):
        with pytest.raises(
            ValueError,
            match=f"The molecule type {molecule_name} is already present in this system.",
        ):
            combined_system.add_molecule_type(
                combined_system.molecule_types[molecule_name],
                1,
            )

    @needs_gmx
    @pytest.mark.filterwarnings("ignore:Automatically up-converting BondHandler from version 0.3 to 0.4.")
    def test_different_force_field_different_energies(
        self,
        combined_system,
        system1,
        molecule2,
        sage,
        parsley,
    ):
        box = Quantity(5 * numpy.eye(3), "nanometer")

        parsley_system = deepcopy(system1)
        sage_system = deepcopy(system1)

        molecule2_parsley = _convert(
            Interchange.from_smirnoff(parsley, [molecule2], box=box),
        )
        molecule2_sage = _convert(Interchange.from_smirnoff(sage, [molecule2], box=box))

        parsley_system.add_molecule_type(molecule2_parsley.molecule_types["MOL__2"], 1)
        sage_system.add_molecule_type(molecule2_sage.molecule_types["MOL__2"], 1)

        parsley_system.positions = combined_system.positions
        sage_system.positions = combined_system.positions

        parsley_system.to_files(prefix="parsley", decimal=8)
        sage_system.to_files(prefix="sage", decimal=8)

        get_intermol_defaults(periodic=True).write_mdp_file("tmp.mdp")

        _parsley_energy = _process(
            _run_gmx_energy("parsley.top", "parsley.gro", "tmp.mdp"),
        )
        _sage_energy = _process(_run_gmx_energy("sage.top", "sage.gro", "tmp.mdp"))

        assert _parsley_energy != _sage_energy

    @needs_gmx
    @pytest.mark.slow
    def test_molecule_order_independent(self, system1, system2):
        positions1 = numpy.vstack([system1.positions, system2.positions])
        positions2 = numpy.vstack([system2.positions, system1.positions])

        system1.add_molecule_type(system2.molecule_types["MOL__2"], 1)
        system1.positions = positions1

        system2.add_molecule_type(system1.molecule_types["MOL__1"], 1)
        system2.positions = positions2

        system1.to_files(prefix="order1", decimal=8)
        system2.to_files(prefix="order2", decimal=8)

        get_intermol_defaults(periodic=True).write_mdp_file("tmp.mdp")

        _process(_run_gmx_energy("order1.top", "order1.gro", "tmp.mdp")).compare(
            _process(_run_gmx_energy("order2.top", "order2.gro", "tmp.mdp")),
        )

    @pytest.mark.slow
    def test_clashing_atom_types(self, combined_system, system1, system2):
        with pytest.raises(
            ValueError,
            match="The molecule type MOL__1 is already present in this system",
        ):
            combined_system.add_molecule_type(system1.molecule_types["MOL__1"], 1)

        with pytest.raises(
            ValueError,
            match="The molecule type MOL__2 is already present in this system",
        ):
            combined_system.add_molecule_type(system2.molecule_types["MOL__2"], 1)

        with pytest.raises(
            ValueError,
            match="The molecule type MOL__1 is already present in this system",
        ):
            system1.add_molecule_type(system1.molecule_types["MOL__1"], 1)

        with pytest.raises(
            ValueError,
            match="The molecule type MOL__2 is already present in this system",
        ):
            system2.add_molecule_type(system2.molecule_types["MOL__2"], 1)


class TestToFiles:
    @needs_gmx
    def test_identical_outputs(self, system1):
        system1.to_files(prefix="1", decimal=8)

        system1.to_top("2.top")
        system1.to_gro("2.gro", decimal=8)

        get_intermol_defaults(periodic=True).write_mdp_file("tmp.mdp")

        _process(_run_gmx_energy("1.top", "1.gro", "tmp.mdp")).compare(
            _process(_run_gmx_energy("2.top", "2.gro", "tmp.mdp")),
        )



from openff.toolkit import Topology
from openff.toolkit.topology._mm_molecule import _SimpleMolecule

from openff.interchange.interop.gromacs._import._import import GROMACSSystem


def _create_topology_from_system(system: GROMACSSystem) -> Topology:
    pass

    topology = Topology()

    for molecule_name, n_copies in system.molecules:
        gromacs_molecule = system.molecule_types[molecule_name]

        molecule = _SimpleMolecule()
        molecule.name = molecule_name

        for atom in gromacs_molecule.atoms:
            atom_type = atom.atom_type
            atomic_number = system.atom_types[atom_type].atomic_number

            molecule.add_atom(atomic_number=atomic_number)

        for bond in gromacs_molecule.bonds:
            molecule.add_bond(bond.atom1 - 1, bond.atom2 - 1)

        for settle in gromacs_molecule.settles:
            # Assume this is a water, ordered as O H H
            assert len(gromacs_molecule.atoms) == 3
            assert molecule.atoms[0].atomic_number == 8
            assert molecule.atoms[1].atomic_number == 1
            assert molecule.atoms[2].atomic_number == 1

            # How to handle H O H? The documentation is unclear about what the first
            # value index must refer to - oxygen or just whatever the first atom is?
            #  > The [ settles ] directive defines the first atom of the water molecule.
            #   https://manual.gromacs.org/current/reference-manual/topologies/constraint-algorithm-section.html#constraintalg
            assert settle.first_atom == 1

            molecule.add_bond(molecule.atom(0), molecule.atom(1))
            molecule.add_bond(molecule.atom(0), molecule.atom(2))

        for _ in range(n_copies):
            topology.add_molecule(molecule)

        # TODO: Add residue metadata

    return topology



from typing import Any, TypedDict

class _AtomTypeInfo(TypedDict):
    whitelist: set[str]
    blacklist: set[str]
    atomtype: str

def find_atomtypes(
    structure: Any,
    forcefield: Any,
    max_iter: int = 10,
) -> dict[int, _AtomTypeInfo]: ...



import numpy
import pytest
from openff.toolkit import ForceField, Molecule, Quantity
from openff.utilities import (
    get_data_file_path,
    has_executable,
    has_package,
    skip_if_missing,
)

from openff.interchange import Interchange
from openff.interchange._tests import get_protein
from openff.interchange.drivers import get_amber_energies, get_openmm_energies

if has_package("openmm"):
    import openmm
    import openmm.app
    import openmm.unit


class TestAmber:
    @pytest.mark.skip(reason="Need replacement route to reference positions")
    def test_inpcrd(self, sage):
        parmed = pytest.importorskip("parmed")

        mol = Molecule.from_smiles(10 * "C")
        mol.name = "HPER"
        mol.generate_conformers(n_conformers=1)

        out = Interchange.from_smirnoff(force_field=sage, topology=mol.to_topology())
        out.box = Quantity([4, 4, 4], "nanometer")
        out.positions = mol.conformers[0]
        out.positions = Quantity(numpy.round(out.positions.m_as("nanometer"), 5), "nanometer")

        out.to_inpcrd("internal.inpcrd")
        # This method no longer exists
        out._to_parmed().save("parmed.inpcrd")

        coords1 = parmed.load_file("internal.inpcrd").coordinates
        coords2 = parmed.load_file("parmed.inpcrd").coordinates

        numpy.testing.assert_equal(coords1, coords2)

    @skip_if_missing("openmm")
    @pytest.mark.skipif(not has_executable("sander"), reason="sander not installed")
    @pytest.mark.parametrize(
        "smiles",
        [
            "C",
            "CC",  # Adds a proper torsion term(s)
            "C=O",  # Simplest molecule with any improper torsion
            "OC=O",  # Simplest molecule with a multi-term torsion
            "CCOC",  # This hits t86, which has a non-1.0 idivf
            "C1COC(=O)O1",  # This adds an improper, i2
        ],
    )
    def test_amber_energy(self, sage_unconstrained, smiles):
        """
        Basic test to see if the amber energy driver is functional.

        Note this test can only use the unconstrained version of Sage because sander applies SHAKE
        constraints in the single-point energy calculation, i.e. uses geometries with constraints
        applied, NOT what is in the coordinate file. See issue
        https://github.com/openforcefield/openff-interchange/issues/323
        """
        mol = Molecule.from_smiles(smiles)
        mol.generate_conformers(n_conformers=1)
        top = mol.to_topology()

        interchange = Interchange.from_smirnoff(sage_unconstrained, top)

        interchange.box = [5, 5, 5]
        interchange.positions = mol.conformers[0]

        omm_energies = get_openmm_energies(interchange, combine_nonbonded_forces=True)
        amb_energies = get_amber_energies(interchange)

        # TODO: More investigation into possible non-bonded energy differences and better reporting.
        #       03/02/2023 manually inspected some files and charges and vdW parameters are
        #       precisely identical. Passing box vectors to prmtop files might not always work.
        omm_energies.energies.pop("Nonbonded")
        amb_energies.energies.pop("vdW")
        amb_energies.energies.pop("Electrostatics")

        omm_energies.compare(amb_energies)


class TestPRMTOP:
    @skip_if_missing("mdtraj")
    @skip_if_missing("MDAnalysis")
    @skip_if_missing("openmm")
    @pytest.mark.slow
    def test_atom_names_pdb(self):
        import MDAnalysis
        import mdtraj

        peptide = get_protein("MainChain_ALA_ALA")
        ff14sb = ForceField("ff14sb_off_impropers_0.0.3.offxml")

        Interchange.from_smirnoff(ff14sb, peptide.to_topology()).to_prmtop(
            "atom_names.prmtop",
        )

        pdb_object = openmm.app.PDBFile(
            get_data_file_path("proteins/MainChain_ALA_ALA.pdb", "openff.toolkit"),
        )
        openmm_object = openmm.app.AmberPrmtopFile("atom_names.prmtop")
        mdanalysis_object = MDAnalysis.Universe("atom_names.prmtop")
        # This may not be useful, see
        # https://github.com/mdtraj/mdtraj/blob/6bb35a8d78a5758ff1f72b3af1fc21d2e38f1029/mdtraj/formats/prmtop.py#L47-L49
        mdtraj_object = mdtraj.load_prmtop("atom_names.prmtop")

        pdb_atom_names = [atom.name for atom in pdb_object.topology.atoms()]

        openmm_atom_names = [atom.name for atom in openmm_object.topology.atoms()]
        mdanalysis_atom_names = [atom.name for atom in mdanalysis_object.atoms]
        mdtraj_atom_names = [atom.name for atom in mdtraj_object.atoms]

        assert openmm_atom_names == pdb_atom_names
        assert mdanalysis_atom_names == pdb_atom_names
        assert mdtraj_atom_names == pdb_atom_names


class TestAmberResidues:
    @pytest.mark.parametrize("patch_residue_name", [True, False])
    def test_single_residue_system_residue_name(
        self,
        tmp_path,
        sage,
        ethanol,
        patch_residue_name,
    ):
        parmed = pytest.importorskip("parmed")
        if patch_residue_name:
            for atom in ethanol.atoms:
                atom.metadata["residue_name"] = "YUP"

            ethanol.add_default_hierarchy_schemes()

        Interchange.from_smirnoff(sage, [ethanol]).to_prmtop("test.prmtop")

        residue_names = [r.name for r in parmed.load_file("test.prmtop").residues]

        if patch_residue_name:
            assert residue_names == ["YUP"]
        else:
            assert residue_names == ["RES"]



import networkx
from openff.toolkit import Quantity, Topology

from openff.interchange import Interchange
from openff.interchange._experimental import experimental
from openff.interchange.common._nonbonded import ElectrostaticsCollection, vdWCollection
from openff.interchange.common._valence import (
    AngleCollection,
    BondCollection,
    ConstraintCollection,
    ImproperTorsionCollection,
    ProperTorsionCollection,
    RyckaertBellemansTorsionCollection,
)
from openff.interchange.components.potentials import Potential
from openff.interchange.interop.gromacs.models.models import (
    GROMACSSystem,
    PeriodicImproperDihedral,
    PeriodicProperDihedral,
    RyckaertBellemansDihedral,
)
from openff.interchange.models import (
    AngleKey,
    BondKey,
    ImproperTorsionKey,
    PotentialKey,
    ProperTorsionKey,
    TopologyKey,
)


@experimental
def to_interchange(
    system: GROMACSSystem,
) -> Interchange:
    """
    Convert a GROMACS system to an Interchange object.

    .. warning :: This method is experimental.

    Parameters
    ----------
    system
        The GROMACS system to convert.

    Returns
    -------
    interchange
        The converted Interchange object.

    """
    if (system.nonbonded_function, system.gen_pairs) != (1, True):
        raise NotImplementedError()

    if system.combination_rule == 2:
        mixing_rule = "lorentz-berthelot"
    elif system.combination_rule == 3:
        mixing_rule = "geometric"

    vdw = vdWCollection(scale14=system.vdw_14, mixing_rule=mixing_rule)
    electrostatics = ElectrostaticsCollection(version=0.4, scale_14=system.coul_14)

    constraints = ConstraintCollection()
    bonds = BondCollection()
    angles = AngleCollection()
    # TODO: Split out periodic and RB/other styles?
    periodic_propers = ProperTorsionCollection()
    rb_torsions = RyckaertBellemansTorsionCollection()
    impropers = ImproperTorsionCollection()

    vdw.potentials = {
        PotentialKey(
            id=f"{atom_type.name}",
            associated_handler="ExternalSource",
        ): Potential(
            parameters={"sigma": atom_type.sigma, "epsilon": atom_type.epsilon},
        )
        for atom_type in system.atom_types.values()
    }

    molecule_start_index = 0

    for molecule_name, n_copies in system.molecules:
        molecule_type = system.molecule_types[molecule_name]

        for _ in range(n_copies):
            for atom in molecule_type.atoms:
                topology_atom_index = molecule_start_index + atom.index - 1
                topology_key = TopologyKey(
                    atom_indices=(topology_atom_index,),
                )

                vdw.key_map.update(
                    {
                        topology_key: PotentialKey(
                            id=f"{atom.atom_type}",
                            associated_handler="ExternalSource",
                        ),
                    },
                )

                # GROMACS does NOT necessarily tie partial charges to atom types, so need a new key for each atom
                electrostatics_key = PotentialKey(
                    id=f"{topology_key.atom_indices[0]}",
                    associated_handler="ExternalSource",
                )
                electrostatics.key_map.update(
                    {topology_key: electrostatics_key},
                )

                electrostatics.potentials.update(
                    {electrostatics_key: Potential(parameters={"charge": atom.charge})},
                )

            if len(molecule_type.settles) > 0:
                if [system.atom_types[a.atom_type].atomic_number for a in molecule_type.atoms] != [
                    8,
                    1,
                    1,
                ]:
                    raise NotImplementedError(
                        "Settles have only been implemented for water with OHH ordering.",
                    )

            for settle in molecule_type.settles:
                oxygen_hydrogen1 = BondKey(
                    atom_indices=(
                        1 + molecule_start_index - 1,
                        2 + molecule_start_index - 1,
                    ),
                )
                oxygen_hydrogen2 = BondKey(
                    atom_indices=(
                        1 + molecule_start_index - 1,
                        3 + molecule_start_index - 1,
                    ),
                )
                hydrogen_hydrogen = BondKey(
                    atom_indices=(
                        2 + molecule_start_index - 1,
                        3 + molecule_start_index - 1,
                    ),
                )

                # Could probably look up the key and potential instead of re-creating?
                oxygen_hydrogen_key = PotentialKey(
                    id="O-H-settles",
                    associated_handler="Constraints",
                )

                hydrogen_hydrogen_key = PotentialKey(
                    id="H-H-settles",
                    associated_handler="Constraints",
                )

                oxygen_hydrogen_constraint = Potential(
                    parameters={
                        "distance": settle.oxygen_hydrogen_distance,
                    },
                )
                hydrogen_hydrogen_constraint = Potential(
                    parameters={
                        "distance": settle.hydrogen_hydrogen_distance,
                    },
                )

                constraints.key_map.update({oxygen_hydrogen1: oxygen_hydrogen_key})
                constraints.key_map.update({oxygen_hydrogen2: oxygen_hydrogen_key})
                constraints.key_map.update({hydrogen_hydrogen: hydrogen_hydrogen_key})

                constraints.potentials.update(
                    {oxygen_hydrogen_key: oxygen_hydrogen_constraint},
                )
                constraints.potentials.update(
                    {hydrogen_hydrogen_key: hydrogen_hydrogen_constraint},
                )

            # TODO: Build constraints from settles
            for bond in molecule_type.bonds:
                topology_key = BondKey(
                    atom_indices=(
                        bond.atom1 + molecule_start_index - 1,
                        bond.atom2 + molecule_start_index - 1,
                    ),
                )

                potential_key = PotentialKey(
                    id="-".join(map(str, topology_key.atom_indices)),
                    associated_handler="ExternalSource",
                )

                potential = Potential(
                    parameters={
                        "k": bond.k,
                        "length": bond.length,
                    },
                )

                bonds.key_map.update({topology_key: potential_key})
                bonds.potentials.update({potential_key: potential})

            for angle in molecule_type.angles:
                topology_key = AngleKey(
                    atom_indices=(
                        angle.atom1 + molecule_start_index - 1,
                        angle.atom2 + molecule_start_index - 1,
                        angle.atom3 + molecule_start_index - 1,
                    ),
                )

                potential_key = PotentialKey(
                    id="-".join(map(str, topology_key.atom_indices)),
                    associated_handler="ExternalSource",
                )

                potential = Potential(
                    parameters={
                        "k": angle.k,
                        "angle": angle.angle,
                    },
                )

                angles.key_map.update({topology_key: potential_key})
                angles.potentials.update({potential_key: potential})

            for dihedral in molecule_type.dihedrals:
                if isinstance(dihedral, PeriodicProperDihedral):
                    key_type = ProperTorsionKey
                    collection = periodic_propers
                elif isinstance(dihedral, RyckaertBellemansDihedral):
                    key_type = ProperTorsionKey
                    collection = rb_torsions  # type: ignore[assignment]
                elif isinstance(dihedral, PeriodicImproperDihedral):
                    key_type = ImproperTorsionKey
                    collection = impropers  # type: ignore[assignment]
                else:
                    raise NotImplementedError(
                        f"Dihedral type {type(dihedral)} not implemented.",
                    )

                _key_assigned = False
                mult = 0
                while not _key_assigned:
                    topology_key = key_type(
                        atom_indices=(
                            dihedral.atom1 + molecule_start_index - 1,
                            dihedral.atom2 + molecule_start_index - 1,
                            dihedral.atom3 + molecule_start_index - 1,
                            dihedral.atom4 + molecule_start_index - 1,
                        ),
                        mult=mult,
                    )
                    if topology_key not in collection.key_map:
                        _key_assigned = True
                    else:
                        mult += 1

                potential_key = PotentialKey(
                    id="-".join(map(str, topology_key.atom_indices)),
                    mult=topology_key.mult,  # type: ignore[attr-defined]
                    associated_handler="ExternalSource",
                )

                if isinstance(
                    dihedral,
                    PeriodicProperDihedral | PeriodicImproperDihedral,
                ):
                    potential = Potential(
                        parameters={
                            "periodicity": Quantity(dihedral.multiplicity, "dimensionless"),
                            "phase": dihedral.phi,
                            "k": dihedral.k,
                            "idivf": Quantity(1, "dimensionless"),
                        },
                    )

                elif isinstance(dihedral, RyckaertBellemansDihedral):
                    potential = Potential(
                        parameters={
                            "c0": dihedral.c0,
                            "c1": dihedral.c1,
                            "c2": dihedral.c2,
                            "c3": dihedral.c3,
                            "c4": dihedral.c4,
                            "c5": dihedral.c5,
                        },
                    )

                else:
                    raise NotImplementedError()

                collection.key_map.update({topology_key: potential_key})
                collection.potentials.update({potential_key: potential})

            molecule_start_index += len(molecule_type.atoms)

    interchange = Interchange(topology=_convert_topology(system))

    interchange.collections.update(
        {
            "vdW": vdw,
            "Electrostatics": electrostatics,
            "Constraints": constraints,
            "Bonds": bonds,
            "Angles": angles,
            "ProperTorsions": periodic_propers,
            "RBTorsions": rb_torsions,
            "ImproperTorsions": impropers,
        },
    )

    interchange.positions = system.positions
    interchange.box = system.box

    return interchange


def _convert_topology(
    system: GROMACSSystem,
) -> Topology:
    from openff.toolkit.topology._mm_molecule import _SimpleMolecule

    topology = Topology()

    for molecule_name, n_molecules in system.molecules:
        molecule_type = system.molecule_types[molecule_name]

        graph = networkx.Graph()

        for atom in molecule_type.atoms:
            graph.add_node(
                atom.index - 1,
                atomic_number=system.atom_types[atom.atom_type].atomic_number,
            )

        if [system.atom_types[atom.atom_type].atomic_number for atom in molecule_type.atoms] == [
            8,
            1,
            1,
        ]:
            graph.add_edge(0, 1)
            graph.add_edge(0, 2)
        else:
            for bond in molecule_type.bonds:
                graph.add_edge(
                    bond.atom1 - 1,
                    bond.atom2 - 1,
                )

        molecule = _SimpleMolecule._from_subgraph(graph)
        molecule.name = molecule_name

        for _ in range(n_molecules):
            topology.add_molecule(molecule)

    return topology



import warnings

from openff.toolkit import ForceField, Molecule, Quantity, Topology
from openff.toolkit.typing.engines.smirnoff import AngleHandler, BondHandler, ParameterHandler, ProperTorsionHandler
from openff.toolkit.typing.engines.smirnoff.plugins import load_handler_plugins
from packaging.version import Version

from openff.interchange import Interchange
from openff.interchange.common._positions import _infer_positions
from openff.interchange.components.toolkit import _check_electrostatics_handlers
from openff.interchange.exceptions import (
    InvalidTopologyError,
    MissingParameterHandlerError,
    PresetChargesError,
    SMIRNOFFHandlersNotImplementedError,
    UnsupportedExportError,
)
from openff.interchange.plugins import load_smirnoff_plugins
from openff.interchange.smirnoff._base import SMIRNOFFCollection, _check_all_valence_terms_assigned
from openff.interchange.smirnoff._gbsa import SMIRNOFFGBSACollection
from openff.interchange.smirnoff._nonbonded import (
    SMIRNOFFElectrostaticsCollection,
    SMIRNOFFvdWCollection,
    _upconvert_vdw_handler,
)
from openff.interchange.smirnoff._valence import (
    SMIRNOFFAngleCollection,
    SMIRNOFFBondCollection,
    SMIRNOFFConstraintCollection,
    SMIRNOFFImproperTorsionCollection,
    SMIRNOFFProperTorsionCollection,
    _upconvert_bondhandler,
)
from openff.interchange.smirnoff._virtual_sites import SMIRNOFFVirtualSiteCollection
from openff.interchange.warnings import PresetChargesAndVirtualSitesWarning

_SUPPORTED_PARAMETER_HANDLERS: set[str] = {
    "Constraints",
    "Bonds",
    "Angles",
    "ProperTorsions",
    "ImproperTorsions",
    "vdW",
    "Electrostatics",
    "LibraryCharges",
    "ChargeIncrementModel",
    "VirtualSites",
    "GBSA",
}

_PLUGIN_CLASS_MAPPING: dict[
    type["ParameterHandler"],
    type["SMIRNOFFCollection"],
] = dict()

for collection_plugin in load_smirnoff_plugins():
    for parameter_handler in collection_plugin.allowed_parameter_handlers():
        if parameter_handler in load_handler_plugins():
            _SUPPORTED_PARAMETER_HANDLERS.add(parameter_handler._TAGNAME)
            _PLUGIN_CLASS_MAPPING[parameter_handler] = collection_plugin
        else:
            raise ValueError(
                f"`SMIRNOFFCollection` plugin {collection_plugin} supports `ParameterHandler` "
                f"plugin {parameter_handler}, but it was not found in the `openff.toolkit.plugins` "
                "entry point. If this collection can use this handler but does not require it, "
                "please raise an issue on GitHub describing your use case.",
            )


def _check_supported_handlers(force_field: ForceField):
    unsupported = list()

    for handler_name in force_field.registered_parameter_handlers:
        if handler_name in {"ToolkitAM1BCC", "NAGLCharges"}:
            continue
        if handler_name not in _SUPPORTED_PARAMETER_HANDLERS:
            unsupported.append(handler_name)

    if unsupported:
        raise SMIRNOFFHandlersNotImplementedError(
            f"SMIRNOFF section(s) not implemented in Interchange: {unsupported}",
        )


def validate_topology(value):
    """Validate a topology-like argument, spliced from a previous validator."""
    if value is None:
        return None
    if isinstance(value, Topology):
        return Topology(other=value)
    elif isinstance(value, list):
        return Topology.from_molecules(value)
    else:
        raise InvalidTopologyError(
            "Could not process topology argument, expected openff.toolkit.Topology. "
            f"Found object of type {type(value)}.",
        )


def _preprocess_preset_charges(
    molecules_with_preset_charges: list[Molecule] | None,
) -> list[Molecule] | None:
    """
    Pre-process the molecules_with_preset_charges argument.

    If molecules_with_preset_charges is None, return None.

    If molecules_with_preset_charges is list[Molecule], ensure that

    1. The input is a list of Molecules
    2. Each molecule has assign partial charges
    3. Ensure no molecules are isomorphic with another in the list

    """
    if molecules_with_preset_charges is None:
        return None

    # This relies on Molecule.__eq__(), which may change and/or not have the same equality criteria
    # as we want here.
    # See https://github.com/openforcefield/openff-interchange/pull/1070#discussion_r1792728179
    molecule_set = {molecule for molecule in molecules_with_preset_charges}

    if len(molecule_set) != len(molecules_with_preset_charges):
        raise PresetChargesError(
            "All molecules in the molecules_with_preset_charges list must be isomorphically unique from each other",
        )

    for molecule in molecules_with_preset_charges:
        if molecule.partial_charges is None:
            raise PresetChargesError(
                "All molecules in the molecules_with_preset_charges list must have partial charges assigned.",
            )

    return molecules_with_preset_charges


def _clear_smirnoff_creation_caches():
    """
    Due to bugs in the toolkit, these caches can be invalid if a force field is modified and then
    sent to (SMIRNOFF) creation pathways again. This function clears those caches - call it whenever
    `Interchange.from_smirnoff` may be called multiple times on the same force field object
    (including possible modifications).

    https://github.com/openforcefield/openff-toolkit/issues/2065
    """
    from openff.interchange.components.toolkit import _cache_angle_parameter_lookup, _cache_torsion_parameter_lookup

    for obj in [
        _cache_angle_parameter_lookup,
        _cache_torsion_parameter_lookup,
    ]:
        obj.cache_clear()


def _create_interchange(
    force_field: ForceField,
    topology: Topology | list[Molecule],
    box: Quantity | None = None,
    positions: Quantity | None = None,
    molecules_with_preset_charges: list[Molecule] | None = None,
    partial_bond_orders_from_molecules: list[Molecule] | None = None,
    allow_nonintegral_charges: bool = False,
) -> Interchange:
    molecules_with_preset_charges = _preprocess_preset_charges(molecules_with_preset_charges)

    _check_supported_handlers(force_field)

    _clear_smirnoff_creation_caches()

    if molecules_with_preset_charges and "VirtualSites" in force_field.registered_parameter_handlers:
        warnings.warn(
            "Preset charges were provided (via `charge_from_molecules`) alongside a force field that includes "
            "virtual site parameters. Note that virtual sites will be applied charges from the force field and "
            "cannot be given preset charges. Virtual sites may also affect the charges of their orientation "
            "atoms, even if those atoms are given preset charges.",
            PresetChargesAndVirtualSitesWarning,
        )

    # interchange = Interchange(topology=topology)
    # or maybe
    interchange = Interchange(topology=validate_topology(topology))

    interchange.positions = _infer_positions(interchange.topology, positions)

    interchange.box = interchange.topology.box_vectors if box is None else box

    _bonds(
        interchange,
        force_field,
        interchange.topology,
        partial_bond_orders_from_molecules,
    )
    _constraints(
        interchange,
        force_field,
        interchange.topology,
        bonds=interchange.collections.get("Bonds", None),  # type: ignore[arg-type]
    )
    if "Bonds" in force_field.registered_parameter_handlers:
        assigned_bond_indices = {tuple(key.atom_indices) for key in interchange["Bonds"].key_map}

        # need to filter the list of _all_ constraints to those that are bond-like, since constraints
        # can also be used to freeze i-k atoms in angles and other things, and we don't want those to
        # be interpreted as over-assigned bond terms
        topological_bond_indices = {
            (
                interchange.topology.atom_index(bond.atom1),
                interchange.topology.atom_index(bond.atom2),
            )
            for bond in interchange.topology.bonds
        }
        all_assigned_constraint_indices = {tuple(key.atom_indices) for key in interchange["Constraints"].key_map}
        bond_like_assigned_constraint_indices = {
            constraint for constraint in all_assigned_constraint_indices if constraint in topological_bond_indices
        }

        _check_all_valence_terms_assigned(
            handler_class=BondHandler,
            topology=interchange.topology,
            assigned_atom_indices=assigned_bond_indices.union(bond_like_assigned_constraint_indices),
            valence_terms=interchange["Bonds"].valence_terms(interchange.topology),
        )

    _angles(interchange, force_field, interchange.topology)

    if "Angles" in force_field.registered_parameter_handlers:
        assigned_angle_indices = {tuple(key.atom_indices) for key in interchange["Angles"].key_map}

        # need to filter the list of angles whose i-k atoms are convered by constraints
        topological_angle_indices = {
            (
                interchange.topology.atom_index(angle[0]),
                interchange.topology.atom_index(angle[1]),
                interchange.topology.atom_index(angle[2]),
            )
            for angle in interchange.topology.angles
        }
        all_assigned_constraint_indices = {tuple(key.atom_indices) for key in interchange["Constraints"].key_map}
        angles_mimicked_by_constraints = {
            pair for pair in topological_angle_indices if tuple((pair[0], pair[2])) in all_assigned_constraint_indices
        }

        _check_all_valence_terms_assigned(
            handler_class=AngleHandler,
            topology=interchange.topology,
            assigned_atom_indices=assigned_angle_indices.union(angles_mimicked_by_constraints),
            valence_terms=interchange["Angles"].valence_terms(interchange.topology),
        )

    _propers(
        interchange,
        force_field,
        interchange.topology,
        partial_bond_orders_from_molecules,
    )
    _impropers(interchange, force_field, interchange.topology)

    for handler_name, handler_class in zip(
        ["ProperTorsions"],
        [ProperTorsionHandler],
    ):
        if handler_name in force_field.registered_parameter_handlers:
            _check_all_valence_terms_assigned(
                handler_class=handler_class,
                topology=interchange.topology,
                assigned_atom_indices={
                    *{tuple(key.atom_indices) for key in interchange[handler_name].key_map},
                },
                valence_terms=interchange[handler_name].valence_terms(interchange.topology),  # type: ignore[attr-defined]
            )

    _vdw(interchange, force_field, interchange.topology)
    _electrostatics(
        interchange,
        force_field,
        interchange.topology,
        molecules_with_preset_charges,
        allow_nonintegral_charges,
    )
    _plugins(interchange, force_field, interchange.topology)

    _virtual_sites(interchange, force_field, interchange.topology)

    _gbsa(interchange, force_field, interchange.topology)

    interchange.topology = interchange.topology

    return interchange


def _bonds(
    interchange: Interchange,
    force_field: ForceField,
    _topology: Topology,
    partial_bond_orders_from_molecules: list[Molecule] | None = None,
):
    if "Bonds" not in force_field.registered_parameter_handlers:
        return

    if force_field["Bonds"].version == Version("0.3"):
        _upconvert_bondhandler(force_field["Bonds"])

    interchange.collections.update(
        {
            "Bonds": SMIRNOFFBondCollection.create(
                parameter_handler=force_field["Bonds"],
                topology=_topology,
                partial_bond_orders_from_molecules=partial_bond_orders_from_molecules,
            ),
        },
    )


def _constraints(
    interchange: Interchange,
    force_field: ForceField,
    topology: Topology,
    bonds: SMIRNOFFBondCollection | None = None,
):
    interchange.collections.update(
        {
            "Constraints": SMIRNOFFConstraintCollection.create(
                parameter_handler=[
                    handler
                    for handler in [
                        force_field._parameter_handlers.get("Bonds", None),
                        force_field._parameter_handlers.get("Constraints", None),
                    ]
                    if handler is not None
                ],
                topology=topology,
                bonds=bonds,
            ),
        },
    )


def _angles(interchange, force_field, _topology):
    if "Angles" not in force_field.registered_parameter_handlers:
        return

    interchange.collections.update(
        {
            "Angles": SMIRNOFFAngleCollection.create(
                parameter_handler=force_field["Angles"],
                topology=_topology,
            ),
        },
    )


def _propers(
    interchange,
    force_field,
    _topology,
    partial_bond_orders_from_molecules=None,
):
    if "ProperTorsions" not in force_field.registered_parameter_handlers:
        return

    interchange.collections.update(
        {
            "ProperTorsions": SMIRNOFFProperTorsionCollection.create(
                parameter_handler=force_field["ProperTorsions"],
                topology=_topology,
                partial_bond_orders_from_molecules=partial_bond_orders_from_molecules,
            ),
        },
    )


def _impropers(interchange, force_field, _topology):
    if "ImproperTorsions" not in force_field.registered_parameter_handlers:
        return

    interchange.collections.update(
        {
            "ImproperTorsions": SMIRNOFFImproperTorsionCollection.create(
                parameter_handler=force_field["ImproperTorsions"],
                topology=_topology,
            ),
        },
    )


def _vdw(interchange: Interchange, force_field: ForceField, topology: Topology):
    if "vdW" not in force_field.registered_parameter_handlers:
        return

    # TODO: This modifies a user-supplied argument in-place, might consider
    # deepcopying it somewhere around `Interchange.from_x`

    _upconvert_vdw_handler(force_field["vdW"])

    interchange.collections.update(
        {
            "vdW": SMIRNOFFvdWCollection.create(
                parameter_handler=force_field["vdW"],
                topology=topology,
            ),
        },
    )


def _electrostatics(
    interchange: Interchange,
    force_field: ForceField,
    topology: Topology,
    molecules_with_preset_charges: list[Molecule] | None = None,
    allow_nonintegral_charges: bool = False,
):
    if "Electrostatics" not in force_field.registered_parameter_handlers:
        if _check_electrostatics_handlers(force_field):
            raise MissingParameterHandlerError(
                "Force field contains parameter handler(s) that may assign/modify "
                "partial charges, but no ElectrostaticsHandler was found.",
            )

        else:
            return

    interchange.collections.update(
        {
            "Electrostatics": SMIRNOFFElectrostaticsCollection.create(
                parameter_handler=[
                    handler
                    for handler in [
                        force_field._parameter_handlers.get(name, None)
                        for name in [
                            "Electrostatics",
                            "NAGLCharges",
                            "ChargeIncrementModel",
                            "ToolkitAM1BCC",
                            "LibraryCharges",
                        ]
                    ]
                    if handler is not None
                ],
                topology=topology,
                molecules_with_preset_charges=molecules_with_preset_charges,
                allow_nonintegral_charges=allow_nonintegral_charges,
            ),
        },
    )


def _gbsa(
    interchange: Interchange,
    force_field: ForceField,
    _topology: Topology,
):
    if "GBSA" not in force_field.registered_parameter_handlers:
        return

    interchange.collections.update(
        {
            "GBSA": SMIRNOFFGBSACollection.create(
                parameter_handler=force_field["GBSA"],
                topology=_topology,
            ),
        },
    )


def _virtual_sites(
    interchange: Interchange,
    force_field: ForceField,
    topology: Topology,
):
    if "VirtualSites" not in force_field.registered_parameter_handlers:
        return

    virtual_site_handler = SMIRNOFFVirtualSiteCollection()

    virtual_site_handler.exclusion_policy = force_field["VirtualSites"].exclusion_policy

    virtual_site_handler.store_matches(
        parameter_handler=force_field["VirtualSites"],
        topology=topology,
    )

    try:
        vdw = interchange["vdW"]
    except LookupError:
        # There might not be a handler named "vdW" but there could be a plugin that
        # is directed to act as it
        for collection in interchange.collections.values():
            if collection.is_plugin:
                if collection.acts_as == "vdW":  # type: ignore[attr-defined]
                    vdw = collection  # type: ignore[assignment]
                    break
        else:
            raise UnsupportedExportError(
                "Virtual sites with no vdW handler not currently supported. If this use case is "
                "important to you, please raise an issue describing the functionality you wish to "
                "see.",
            )

    electrostatics: SMIRNOFFElectrostaticsCollection = interchange["Electrostatics"]  # type: ignore[assignment]

    virtual_site_handler.store_potentials(
        parameter_handler=force_field["VirtualSites"],
        vdw_collection=vdw,  # type: ignore[arg-type]
        electrostatics_collection=electrostatics,
    )

    interchange.collections.update({"VirtualSites": virtual_site_handler})


def _build_collection(
    collection_class: type[SMIRNOFFCollection],
    handler_classes: list[type[ParameterHandler]],
    interchange: Interchange,
    force_field: ForceField,
    topology: Topology,
):
    kwargs = {}

    # Always attach handlers
    if len(handler_classes) == 1:
        kwargs["parameter_handler"] = force_field[handler_classes[0]._TAGNAME]
    else:
        kwargs["parameter_handler"] = [force_field[cls._TAGNAME] for cls in handler_classes]

    kwargs["topology"] = topology

    # VirtualSite collections need vdW + Electrostatics collections, and we can identify
    # them by the allowed_vdw_parameter_handlers class method.
    if hasattr(collection_class, "allowed_vdw_parameter_handlers"):
        vdw_tagnames = [x._TAGNAME for x in collection_class.allowed_vdw_parameter_handlers()]

        if (n_vdw_tagnames := len(vdw_tagnames)) != 1:
            raise NotImplementedError(
                f"Collection {collection_class} requires {n_vdw_tagnames} vdW handlers, but only one is supported.",
            )

        kwargs["vdw_collection"] = interchange[vdw_tagnames[0]]
        kwargs["electrostatics_collection"] = interchange["Electrostatics"]

    return collection_class.create(**kwargs)


def _plugins(
    interchange: Interchange,
    force_field: ForceField,
    topology: Topology,
):
    for collection_class in _PLUGIN_CLASS_MAPPING.values():
        # Track the handlers (keys) that map to this collection (value)
        handler_classes = [
            handler for handler in _PLUGIN_CLASS_MAPPING if _PLUGIN_CLASS_MAPPING[handler] == collection_class
        ]

        if not all(
            [handler_class._TAGNAME in force_field.registered_parameter_handlers for handler_class in handler_classes],
        ):
            continue

        if len(handler_classes) == 0:
            continue

        collection = _build_collection(collection_class, handler_classes, interchange, force_field, topology)

        # No matter if this collection takes one or multiple handlers, key it by its own name
        interchange.collections[collection.type] = collection



import copy
import random
from collections import defaultdict

import numpy
import openmm
import pytest
from openff.toolkit import ForceField, Molecule, Quantity, Topology
from openff.utilities import get_data_file_path, has_package, skip_if_missing

from openff.interchange import Interchange
from openff.interchange._tests import MoleculeWithConformer, needs_gmx
from openff.interchange.constants import kj_mol
from openff.interchange.drivers.openmm import get_openmm_energies
from openff.interchange.exceptions import UnsupportedImportError
from openff.interchange.interop.openmm._import._import import _convert_nonbonded_force
from openff.interchange.warnings import MissingPositionsWarning

if has_package("openmm"):
    import openmm.app
    import openmm.unit


@skip_if_missing("openmm")
class TestFromOpenMM:
    def test_simple_roundtrip(self, sage_unconstrained, ethanol):
        ethanol.generate_conformers(n_conformers=1)

        interchange = Interchange.from_smirnoff(
            sage_unconstrained,
            [ethanol],
            box=Quantity([4, 4, 4], "nanometer"),
        )

        system = interchange.to_openmm(combine_nonbonded_forces=True)

        converted = Interchange.from_openmm(
            system=system,
            topology=interchange.topology.to_openmm(),
            positions=interchange.positions,
            box_vectors=interchange.box,
        )

        get_openmm_energies(interchange).compare(
            get_openmm_energies(converted),
            tolerances={
                "Bond": 1e-6 * kj_mol,
                "Angle": 1e-6 * kj_mol,
                "Torsion": 1e-6 * kj_mol,
                "Nonbonded": 1e-3 * kj_mol,
            },
        )

        assert isinstance(converted.box.m, numpy.ndarray)

        # OpenMM seems to avoid using the built-in type
        assert converted.box.m.dtype in (float, numpy.float32, numpy.float64)

    @pytest.fixture
    def simple_system(self):
        return openmm.XmlSerializer.deserialize(
            open(
                get_data_file_path(
                    "system.xml",
                    "openff.interchange._tests.data",
                ),
            ).read(),
        )

    @pytest.mark.parametrize("as_argument", [False, True])
    def test_different_ways_to_process_box_vectors(
        self,
        as_argument,
        simple_system,
    ):
        topology = Molecule.from_smiles("C").to_topology()
        topology._molecule_virtual_site_map = defaultdict(list)
        topology._particle_map = {index: index for index in range(topology.n_atoms)}

        if as_argument:
            box = Interchange.from_openmm(
                system=simple_system,
                topology=topology,
                box_vectors=simple_system.getDefaultPeriodicBoxVectors(),
            ).box
        else:
            box = Interchange.from_openmm(
                system=simple_system,
                topology=topology,
            ).box

        assert box.shape == (3, 3)

        assert type(box.m[2][2]) in (float, numpy.float64, numpy.float32)
        assert type(box.m[1][1]) is not Quantity

    def test_topology_and_system_box_vectors_differ(
        self,
        simple_system,
    ):
        """Ensure that, if box vectors specified in the topology and system differ, those in the topology are used."""
        topology = Molecule.from_smiles("C").to_topology()
        topology.box_vectors = Quantity([4, 5, 6], "nanometer")

        box = Interchange.from_openmm(
            system=simple_system,
            topology=topology.to_openmm(),
        ).box

        assert numpy.diag(box.m_as("nanometer")) == pytest.approx([4, 5, 6])

    def test_openmm_roundtrip_metadata(self, sage):
        # Make an example OpenMM Topology with metadata.
        # Here we use OFFTK to make the OpenMM Topology, but this could just as easily come from another source
        ethanol = Molecule.from_smiles("CCO")
        benzene = Molecule.from_smiles("c1ccccc1")

        for atom in ethanol.atoms:
            atom.metadata["chain_id"] = "1"
            atom.metadata["residue_number"] = "1"
            atom.metadata["insertion_code"] = ""
            atom.metadata["residue_name"] = "ETH"
        for atom in benzene.atoms:
            atom.metadata["chain_id"] = "1"
            atom.metadata["residue_number"] = "2"
            atom.metadata["insertion_code"] = "A"
            atom.metadata["residue_name"] = "BNZ"

        topology = Topology.from_molecules([ethanol, benzene])
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        # Roundtrip the topology with metadata through openmm,
        # which requires a system even though it won't be used here
        # and requires PME for now
        interchange = Interchange.from_openmm(
            system=sage.create_openmm_system(topology),
            topology=topology.to_openmm(),
        )

        # Ensure that the metadata is the same
        for atom in interchange.topology.molecule(0).atoms:
            assert atom.metadata["chain_id"] == "1"
            assert atom.metadata["residue_number"] == "1"
            assert atom.metadata["insertion_code"] == ""
            assert atom.metadata["residue_name"] == "ETH"
        for atom in interchange.topology.molecule(1).atoms:
            assert atom.metadata["chain_id"] == "1"
            assert atom.metadata["residue_number"] == "2"
            assert atom.metadata["insertion_code"] == "A"
            assert atom.metadata["residue_name"] == "BNZ"

    @pytest.mark.slow
    def test_openmm_native_roundtrip_metadata(self, sage):
        """
        Test that metadata is the same whether we load a PDB through OpenMM+Interchange vs. Topology.from_pdb.
        """
        pdb = openmm.app.PDBFile(
            get_data_file_path(
                "ALA_GLY/ALA_GLY.pdb",
                "openff.interchange._tests.data",
            ),
        )

        topology = Topology.from_pdb(
            get_data_file_path("ALA_GLY/ALA_GLY.pdb", "openff.interchange._tests.data"),
        )

        interchange = Interchange.from_openmm(
            system=sage.create_openmm_system(topology),
            topology=pdb.topology,
        )

        for roundtrip_atom, off_atom in zip(interchange.topology.atoms, topology.atoms):
            # off_atom's metadata also includes a little info about how the chemistry was
            # assigned, so we remove this from the comparison
            off_atom_metadata = copy.deepcopy(off_atom.metadata)
            del off_atom_metadata["match_info"]
            assert roundtrip_atom.metadata == off_atom_metadata

    def test_electrostatics_cutoff_not_ignored(self, ethanol):
        pytest.importorskip("openmmforcefields")

        import openmm.app
        import openmm.unit
        from openmmforcefields.generators import GAFFTemplateGenerator

        topology = ethanol.to_topology()
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        gaff = GAFFTemplateGenerator(molecules=ethanol)
        force_field = openmm.app.ForceField()

        force_field.registerTemplateGenerator(gaff.generator)

        system = force_field.createSystem(
            topology=topology.to_openmm(),
            nonbondedMethod=openmm.app.PME,
            nonbondedCutoff=1.2345 * openmm.unit.nanometer,
        )

        interchange = Interchange.from_openmm(
            system=system,
            topology=topology.to_openmm(),
        )

        assert pytest.approx(1.2345) == interchange["Electrostatics"].cutoff.m_as("nanometer")
        assert pytest.approx(1.2345) == interchange["vdW"].cutoff.m_as("nanometer")

    @needs_gmx
    def test_fill_in_rigid_water_parameters(self, water_dimer):
        import openmm.app

        from openff.interchange.drivers import get_gromacs_energies

        openmm_force_field = openmm.app.ForceField("tip3p.xml")
        openmm_topology = water_dimer.to_openmm()

        imported = Interchange.from_openmm(
            system=openmm_force_field.createSystem(
                openmm_topology,
                constraints=openmm.app.HBonds,
                nonbondedMethod=openmm.app.PME,
            ),
            topology=openmm_topology,
            positions=water_dimer.get_positions().to_openmm(),
        )

        assert len(imported["Bonds"].key_map) == imported.topology.n_bonds == 4
        assert len(imported["Angles"].key_map) == imported.topology.n_angles == 2

        # Mostly just ensure GROMACS evaluation doesn't crash
        get_openmm_energies(imported, combine_nonbonded_forces=False).compare(
            get_gromacs_energies(imported),
            tolerances={
                "Electrostatics": Quantity(0.1, "kilojoule / mole"),
            },
        )

    def test_only_constrained_water(
        self,
        water_dimer,
        default_integrator,
    ):
        water_dimer.box_vectors = Quantity([4, 4, 4], "nanometer")

        interchange = ForceField("openff-2.2.1.offxml").create_interchange(water_dimer)

        simulation = interchange.to_openmm_simulation(integrator=default_integrator)
        system = simulation.system

        for index in range(system.getNumForces()):
            if isinstance(system.getForce(index), openmm.HarmonicBondForce):
                break

        system.removeForce(index)

        for index in range(system.getNumForces()):
            if isinstance(system.getForce(index), openmm.HarmonicAngleForce):
                break

        system.removeForce(index)

        interchange2 = Interchange.from_openmm(
            system=simulation.system,
            topology=simulation.topology,
        )

        assert interchange2.topology.n_bonds == interchange.topology.n_bonds

    def test_use_openmm_topology(self, sage, basic_top):
        Interchange.from_openmm(
            system=sage.create_openmm_system(basic_top),
            topology=basic_top,
        )


@skip_if_missing("openmm")
class TestProcessTopology:
    def test_with_openff_topology(self, sage, basic_top):
        system = sage.create_openmm_system(basic_top)

        basic_top._molecule_virtual_site_map = defaultdict(list)
        basic_top._particle_map = {index: index for index in range(basic_top.n_atoms)}

        with_openff = Interchange.from_openmm(
            system=system,
            topology=basic_top,
        )

        # positions are lost when making an openmm.app.Topology,
        # just feed them back in to make comparison easier
        with_openmm = Interchange.from_openmm(
            system=system,
            topology=basic_top.to_openmm(),
            positions=basic_top.get_positions().to_openmm(),
        )

        assert with_openff.topology.n_atoms == with_openmm.topology.n_atoms
        assert with_openff.topology.n_bonds == with_openmm.topology.n_bonds

        get_openmm_energies(
            with_openff,
            combine_nonbonded_forces=True,
        ).compare(
            get_openmm_energies(with_openmm),
            tolerances={
                "Angle": Quantity(0.001, "kilojoule / mole"),
                "Nonbonded": Quantity(0.001, "kilojoule / mole"),
            },
        )

    @pytest.mark.xfail(reason="Flaky in CI, reliable locally")
    @pytest.mark.parametrize("use_original_topology", [True, False])
    def test_openmm_roundtrip_missing_positions(self, use_original_topology):
        topology = MoleculeWithConformer.from_smiles("CCCCCCO").to_topology()
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        interchange = ForceField("openff-2.2.1.offxml").create_interchange(topology)

        if use_original_topology:
            roundtripped = Interchange.from_openmm(
                system=interchange.to_openmm_system(),
                topology=topology,
            )
        else:
            # need to pass positions if using Interchange.topology since
            # Interchange.topology.get_positions() doesn't work. Split into two test cases

            # when passed no positions (default value), valid use case but warning is thrown
            with pytest.warns(MissingPositionsWarning):
                roundtripped = Interchange.from_openmm(
                    system=interchange.to_openmm_system(),
                    topology=interchange.topology,
                    positions=None,
                )

            assert roundtripped.positions is None

            # when positions **are** passed, they should be used
            roundtripped = Interchange.from_openmm(
                system=interchange.to_openmm_system(),
                topology=interchange.topology,
                positions=interchange.positions,
            )

            assert roundtripped.positions is not None

        assert numpy.allclose(interchange.positions.m_as("nanometer"), topology.get_positions().m_as("nanometer"))
        assert numpy.allclose(roundtripped.positions.m_as("nanometer"), topology.get_positions().m_as("nanometer"))

        get_openmm_energies(interchange, combine_nonbonded_forces=False).compare(
            get_openmm_energies(roundtripped, combine_nonbonded_forces=False),
        )


@skip_if_missing("openmm")
class TestConvertNonbondedForce:
    def test_unsupported_method(self):
        # Cannot parametrize with a class in an optional module
        for method in (
            openmm.NonbondedForce.NoCutoff,
            openmm.NonbondedForce.CutoffNonPeriodic,
            openmm.NonbondedForce.CutoffPeriodic,
        ):
            force = openmm.NonbondedForce()
            force.setNonbondedMethod(method)

            with pytest.raises(UnsupportedImportError):
                _convert_nonbonded_force(force, dict())

    def test_parse_switching_distance(self):
        force = openmm.NonbondedForce()
        force.setNonbondedMethod(openmm.NonbondedForce.PME)

        cutoff = 1 + random.random() * 0.1
        switch_width = random.random() * 0.1

        force.setCutoffDistance(cutoff)
        force.setUseSwitchingFunction(True)
        force.setSwitchingDistance(cutoff - switch_width)

        vdw, _ = _convert_nonbonded_force(force=force, particle_map=dict())

        assert vdw.cutoff.m_as("nanometer") == pytest.approx(cutoff)
        assert vdw.switch_width.m_as("nanometer") == pytest.approx(switch_width)

    def test_parse_switching_distance_unused(self):
        force = openmm.NonbondedForce()
        force.setNonbondedMethod(openmm.NonbondedForce.PME)

        cutoff = 1 + random.random() * 0.1

        force.setCutoffDistance(cutoff)

        vdw, _ = _convert_nonbonded_force(force=force, particle_map=dict())

        assert vdw.cutoff.m_as("nanometer") == pytest.approx(cutoff)
        assert vdw.switch_width.m_as("nanometer") == 0.0


@skip_if_missing("openmm")
class TestConvertConstraints:
    def test_num_constraints(self, sage, basic_top):
        """Test that the number of constraints is preserved when converting to and from OpenMM"""
        interchange = sage.create_interchange(basic_top)

        converted = Interchange.from_openmm(
            topology=interchange.topology.to_openmm(),
            system=interchange.to_openmm(combine_nonbonded_forces=True),
        )

        assert "Constraints" in interchange.collections
        assert "Constraints" in converted.collections

        assert len(interchange["Constraints"].key_map) == len(
            converted["Constraints"].key_map,
        )



"""
Commonly-used constants.
"""

from openff.units import Unit

kj_mol = Unit("kilojoule / mol")
kcal_mol = Unit("kilocalorie_per_mole")

kcal_ang = Unit("kilojoule / mol / angstrom**2")
kcal_rad = Unit("kilojoule / mol / radian**2")

kj_nm = Unit("kilojoule/ mol / nanometer**2")
kj_rad = Unit("kilojoule/ mol / radian**2")

kcal_mol_a2 = Unit("kilocalorie / mol / angstrom**2")
kcal_mol_rad2 = Unit("kilocalorie / mol / radian**2")

_PME = "Ewald3D-ConductingBoundary"
AMBER_COULOMBS_CONSTANT = 18.2223



"""Classes used to represent GROMACS state."""

from typing import Annotated

from openff.toolkit import Quantity
from pydantic import (
    Field,
    PositiveInt,
    PrivateAttr,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    WrapValidator,
)

from openff.interchange._annotations import _DistanceQuantity
from openff.interchange.pydantic import _BaseModel


def validate_particle_type(
    value: str,
    handler: ValidatorFunctionWrapHandler,
    info: ValidationInfo,
) -> str:
    """Validate the particle_type field."""
    # info.data is like the extra values argument in v1
    values = info.data

    if values["mass"].m == 0.0:
        assert value in ("D", "V"), 'Particle type must be "D" or "V" if massless'
    elif values["mass"].m > 0.0:
        assert value == "A", 'Particle type must be "A" if it has mass'

    return value


_ParticleType = Annotated[
    str,
    WrapValidator(validate_particle_type),
]


class GROMACSAtomType(_BaseModel):
    """Base class for GROMACS atom types."""

    name: str
    bonding_type: str = ""
    atomic_number: int
    mass: Quantity
    charge: Quantity
    particle_type: _ParticleType


class LennardJonesAtomType(GROMACSAtomType):
    """A Lennard-Jones atom type."""

    sigma: Quantity
    epsilon: Quantity


class GROMACSAtom(_BaseModel):
    """Base class for GROMACS atoms."""

    index: PositiveInt
    name: str
    atom_type: str  # Maybe this should point to an object
    residue_index: PositiveInt
    residue_name: str
    charge_group_number: PositiveInt
    charge: Quantity
    mass: Quantity


# Should the physical values (distance/angles) be float or Quantity?
class GROMACSVirtualSite(_BaseModel):
    """Base class for storing GROMACS virtual sites."""

    type: str
    name: str
    header_tag: Annotated[int, Field(strict=True, ge=2)]
    site: PositiveInt
    func: PositiveInt
    orientation_atoms: list[int]


class GROMACSVirtualSite2(GROMACSVirtualSite):
    """GROMACS virtual site type 2."""

    type: str = "2"
    header_tag: int = 2
    func: int = 1
    a: float


class GROMACSVirtualSite3(GROMACSVirtualSite):
    """GROMACS virtual site type 3."""

    type: str = "3"
    header_tag: int = 3
    func: int = 1
    a: float
    b: float


class GROMACSVirtualSite3fd(GROMACSVirtualSite):
    """GROMACS virtual site type 3fd."""

    type: str = "3fd"
    header_tag: int = 3
    func: int = 2
    a: float
    d: float


class GROMACSVirtualSite3fad(GROMACSVirtualSite):
    """GROMACS virtual site type 3fad."""

    type: str = "3fad"
    header_tag: int = 3
    func: int = 3
    theta: float
    d: float


class GROMACSVirtualSite3out(GROMACSVirtualSite):
    """GROMACS virtual site type 3out."""

    type: str = "3out"
    header_tag: int = 3
    func: int = 4
    a: float
    b: float
    c: float


class GROMACSVirtualSite4fdn(GROMACSVirtualSite):
    """GROMACS virtual site type 4fdn."""

    type: str = "4fdn"
    header_tag: int = 4
    func: int = 2
    a: float
    b: float
    c: float


class GROMACSBond(_BaseModel):
    """A GROMACS bond."""

    atom1: PositiveInt = Field(
        description="The GROMACS index of the first atom in the bond.",
    )
    atom2: PositiveInt = Field(
        description="The GROMACS index of the second atom in the bond.",
    )
    function: int = Field(1, description="The GROMACS bond function type.")
    length: Quantity
    k: Quantity


class GROMACSPair(_BaseModel):
    """A GROMACS pair."""

    atom1: PositiveInt = Field(
        description="The GROMACS index of the first atom in the pair.",
    )
    atom2: PositiveInt = Field(
        description="The GROMACS index of the second atom in the pair.",
    )


class GROMACSSettles(_BaseModel):
    """A settles-style constraint for water."""

    first_atom: PositiveInt = Field(
        description="The GROMACS index of the first atom in the water.",
    )

    oxygen_hydrogen_distance: _DistanceQuantity = Field(
        description="The fixed distance between the oxygen and hydrogen.",
    )

    hydrogen_hydrogen_distance: _DistanceQuantity = Field(
        description="The fixed distance between the oxygen and hydrogen.",
    )


class GROMACSExclusion(_BaseModel):
    """An Exclusion between an atom and other(s)."""

    # Extra exclusions within a molecule can be added manually in a [ exclusions ] section. Each
    # line should start with one atom index, followed by one or more atom indices. All non-bonded
    # interactions between the first atom and the other atoms will be excluded.

    first_atom: PositiveInt

    other_atoms: list[PositiveInt]


class GROMACSAngle(_BaseModel):
    """A GROMACS angle."""

    atom1: PositiveInt = Field(
        description="The GROMACS index of the first atom in the angle.",
    )
    atom2: PositiveInt = Field(
        description="The GROMACS index of the second atom in the angle.",
    )
    atom3: PositiveInt = Field(
        description="The GROMACS index of the third atom in the angle.",
    )
    angle: Quantity
    k: Quantity


class GROMACSDihedral(_BaseModel):
    """A GROMACS dihedral."""

    atom1: PositiveInt = Field(
        description="The GROMACS index of the first atom in the dihedral.",
    )
    atom2: PositiveInt = Field(
        description="The GROMACS index of the second atom in the dihedral.",
    )
    atom3: PositiveInt = Field(
        description="The GROMACS index of the third atom in the dihedral.",
    )
    atom4: PositiveInt = Field(
        description="The GROMACS index of the fourth atom in the dihedral.",
    )


# TODO: Subclasses could define their allowed "function type" as an extra runtime safeguard?
class PeriodicProperDihedral(GROMACSDihedral):
    """A type 1 dihedral in GROMACS."""

    phi: Quantity
    k: Quantity
    multiplicity: PositiveInt


class RyckaertBellemansDihedral(GROMACSDihedral):
    """A type 3 dihedral in GROMACS."""

    c0: Quantity
    c1: Quantity
    c2: Quantity
    c3: Quantity
    c4: Quantity
    c5: Quantity


class PeriodicImproperDihedral(GROMACSDihedral):
    """A type 4 dihedral in GROMACS."""

    phi: Quantity
    k: Quantity
    multiplicity: PositiveInt


class GROMACSMolecule(_BaseModel):
    """Base class for GROMACS molecules."""

    name: str
    nrexcl: int = Field(
        3,
        description="The farthest neighbor distance whose interactions should be excluded.",
    )

    atoms: list[GROMACSAtom] = Field(
        list(),
        description="The atoms in this molecule.",
    )
    pairs: list[GROMACSPair] = Field(
        list(),
        description="The pairs in this molecule.",
    )
    settles: list[GROMACSSettles] = Field(
        list(),
        description="The settles in this molecule.",
    )
    bonds: list[GROMACSBond] = Field(
        list(),
        description="The bonds in this molecule.",
    )
    angles: list[GROMACSAngle] = Field(
        list(),
        description="The angles in this molecule.",
    )
    dihedrals: list[GROMACSDihedral] = Field(
        list(),
        description="The dihedrals in this molecule.",
    )
    virtual_sites: list[GROMACSVirtualSite] = Field(
        list(),
        description="The virtual sites in this molecule.",
    )
    exclusions: list[GROMACSExclusion] = Field(
        list(),
        description="The exclusions in this molecule.",
    )

    # TODO: This can desync between system- and molecule-level data, it might be better
    #       to not store atom types at the system level, instead storing them at the
    #       molecule and grouping up to system level at write time
    _contained_atom_types: dict[str, LennardJonesAtomType] = PrivateAttr()


class GROMACSSystem(_BaseModel):
    """A GROMACS system. Adapted from Intermol."""

    positions: _DistanceQuantity | None = None
    box: _DistanceQuantity | None = None

    name: str = ""
    nonbonded_function: int = Field(
        1,
        ge=1,
        le=1,
        description="The nonbonded function.",
    )
    combination_rule: int = Field(
        2,
        ge=1,
        le=3,
        description="The combination rule.",
    )
    gen_pairs: bool = Field(True, description="Whether or not to generate pairs.")
    vdw_14: float = Field(
        0.5,
        description="The 1-4 scaling factor for dispersion interactions.",
    )
    coul_14: float = Field(
        0.833333,
        description="The 1-4 scaling factor for electrostatic interactions.",
    )
    atom_types: dict[str, LennardJonesAtomType] = Field(
        dict(),
        description="Atom types, keyed by name.",
    )
    molecule_types: dict[str, GROMACSMolecule] = Field(
        dict(),
        description="Molecule types, keyed by name.",
    )
    molecules: list[tuple[str, int]] = Field(
        list(),
        description="The number of each molecule type in the system, ordered topologically.",
    )

    @classmethod
    def from_files(cls, top_file, gro_file):
        """Parse a GROMACS topology file."""
        from openff.interchange.interop.gromacs._import._import import from_files

        return from_files(top_file, gro_file, cls=cls)

    def to_files(self, prefix: str, decimal: int = 3):
        """Write a GROMACS topology file."""
        from openff.interchange.interop.gromacs.export._export import GROMACSWriter

        writer = GROMACSWriter(
            system=self,
            top_file=f"{prefix}.top",
            gro_file=f"{prefix}.gro",
        )

        writer.to_top()
        writer.to_gro(decimal=decimal)

    def to_top(self, file: str):
        """Write a GROMACS topology file."""
        from openff.interchange.interop.gromacs.export._export import GROMACSWriter

        GROMACSWriter(
            system=self,
            top_file=file,
            gro_file="_.gro",
        ).to_top()

    def to_gro(self, file: str, decimal: int = 3):
        """Write a GROMACS coordinate file."""
        from openff.interchange.interop.gromacs.export._export import GROMACSWriter

        GROMACSWriter(
            system=self,
            top_file="_.gro",
            gro_file=file,
        ).to_gro(decimal=decimal)

    def remove_molecule_type(self, molecule_name: str, n_copies: int = 1):
        """Remove a molecule type from the system."""
        import numpy

        if molecule_name not in self.molecule_types:
            raise ValueError(
                f"The molecule type {molecule_name} is not present in this system.",
            )

        if n_copies != 1 or sum(_n_copies for name, _n_copies in self.molecules if name == molecule_name) != 1:
            raise NotImplementedError()

        n_atoms_before = sum(
            len(self.molecule_types[molecule_name].atoms) * n_copies
            for molecule_name, n_copies in self.molecules[
                : self.molecules.index(
                    (molecule_name, 1),
                )
            ]
        )

        if self.positions is not None:
            row_indices_to_delete = [
                *range(
                    n_atoms_before,
                    n_atoms_before + len(self.molecule_types[molecule_name].atoms),
                ),
            ]

            # Pint lacks __array_function__ needed here, so strip and then tag units
            self.positions = Quantity(
                numpy.delete(self.positions.m, row_indices_to_delete, axis=0),
                self.positions.units,
            )

        self.molecule_types.pop(molecule_name)
        self.molecules = [(name, n_copies) for name, n_copies in self.molecules if name != molecule_name]

    def add_molecule_type(self, molecule: GROMACSMolecule, n_copies: int):
        """Add a molecule type to the system."""
        if molecule.name in self.molecule_types:
            raise ValueError(
                f"The molecule type {molecule.name} is already present in this system.",
            )

        if len(molecule._contained_atom_types) == 0:
            raise ValueError(
                f"The molecule type {molecule.name} does not contain any atom types.",
            )

        for atom_type_name, atom_type in molecule._contained_atom_types.items():
            if atom_type_name in self.atom_types:
                raise ValueError(
                    f"An atom type {atom_type_name} is already present in this system.",
                )

            self.atom_types[atom_type_name] = atom_type

        self.molecule_types[molecule.name] = molecule
        self.molecules.append((molecule.name, n_copies))



"""Test SMIRNOFF-GROMACS conversion."""

import pytest
from openff.toolkit import Molecule

from openff.interchange import Interchange
from openff.interchange.interop.gromacs.models.models import GROMACSMolecule
from openff.interchange.smirnoff._gromacs import (
    _convert,
    _convert_angles,
    _convert_bonds,
    _convert_settles,
)


class TestConvert:
    def test_residue_names(self, sage):
        """Reproduce issue #642."""
        ligand = Molecule.from_smiles("CCO")
        ligand.generate_conformers(n_conformers=1)

        for atom in ligand.atoms:
            atom.metadata["residue_name"] = "LIG"

        system = _convert(
            Interchange.from_smirnoff(
                sage,
                [ligand],
            ),
        )

        for molecule_type in system.molecule_types.values():
            for atom in molecule_type.atoms:
                assert atom.residue_name == "LIG"


class TestSettles:
    @pytest.fixture
    def tip3p_interchange(self, tip3p, water):
        return tip3p.create_interchange(water.to_topology())

    @pytest.fixture
    def sage_tip3p_interchange(self, sage, water):
        return sage.create_interchange(water.to_topology())

    def test_catch_other_water_ordering(self, tip3p):
        molecule = Molecule.from_mapped_smiles("[H:1][O:2][H:3]")
        interchange = tip3p.create_interchange(molecule.to_topology())

        with pytest.raises(Exception, match="OHH"):
            _convert_settles(
                GROMACSMolecule(name="foo"),
                interchange.topology.molecule(0),
                interchange,
            )

    @pytest.mark.parametrize("use_bundled_tip3p", [True, False])
    def test_convert_settles(
        self,
        use_bundled_tip3p,
        tip3p_interchange,
        sage_tip3p_interchange,
    ):
        molecule = GROMACSMolecule(name="foo")

        if use_bundled_tip3p:
            _convert_settles(
                molecule,
                sage_tip3p_interchange.topology.molecule(0),
                sage_tip3p_interchange,
            )
        else:
            _convert_settles(
                molecule,
                tip3p_interchange.topology.molecule(0),
                tip3p_interchange,
            )

        assert len(molecule.settles) == 1

        settle = molecule.settles[0]

        assert settle.first_atom == 1
        assert settle.hydrogen_hydrogen_distance.m_as("angstrom") == pytest.approx(
            1.5139006545247014,
        )
        assert settle.oxygen_hydrogen_distance.m_as("angstrom") == pytest.approx(
            0.9572,
        )

        assert molecule.exclusions[0].first_atom == 1
        assert molecule.exclusions[0].other_atoms == [2, 3]
        assert molecule.exclusions[1].first_atom == 2
        assert molecule.exclusions[1].other_atoms == [1, 3]
        assert molecule.exclusions[2].first_atom == 3
        assert molecule.exclusions[2].other_atoms == [1, 2]

    def test_convert_no_settles_unconstrained_water(self, tip3p_interchange):
        tip3p_interchange.collections["Constraints"].key_map = dict()

        molecule = GROMACSMolecule(name="foo")

        _convert_settles(
            molecule,
            tip3p_interchange.topology.molecule(0),
            tip3p_interchange,
        )

        assert len(molecule.settles) == 0

    def test_convert_no_settles_no_constraints(self, tip3p_interchange):
        tip3p_interchange.collections.pop("Constraints")

        molecule = GROMACSMolecule(name="foo")

        _convert_settles(
            molecule,
            tip3p_interchange.topology.molecule(0),
            tip3p_interchange,
        )

        assert len(molecule.settles) == 0

    @pytest.mark.parametrize("use_bundled_tip3p", [True, False])
    def test_error_if_water_partially_constrained(
        self,
        use_bundled_tip3p,
        tip3p_interchange,
        sage_tip3p_interchange,
    ):
        from openff.interchange.models import BondKey

        # Manually remove the H-H constraint parameter from each,
        tip3p_interchange["Constraints"].potentials.pop(
            tip3p_interchange["Constraints"].key_map[BondKey(atom_indices=(0, 1))],
        )

        sage_tip3p_interchange["Constraints"].potentials.pop(
            sage_tip3p_interchange["Constraints"].key_map[BondKey(atom_indices=(0, 1))],
        )

        molecule = GROMACSMolecule(name="foo")

        if not use_bundled_tip3p:
            with pytest.raises(
                RuntimeError,
                match=r"Could not find a constraint distance .*0.*1",
            ):
                # ... and ensure this is an error when there is no bond parameter
                # to fall back on
                _convert_settles(
                    GROMACSMolecule(name="foo"),
                    tip3p_interchange.topology.molecule(0),
                    tip3p_interchange,
                )
        else:
            # ... but uses a O-H distance if there is one (here, from Sage)
            _convert_settles(
                molecule,
                sage_tip3p_interchange.topology.molecule(0),
                sage_tip3p_interchange,
            )

            # b88 in both 2.0 and 2.1
            assert molecule.settles[0].oxygen_hydrogen_distance.m_as(
                "angstrom",
            ) == pytest.approx(0.971162542)

    def test_no_bonds_or_angles_if_settle(self, tip3p_interchange):
        molecule = GROMACSMolecule(name="foo")

        for function in [_convert_settles, _convert_bonds, _convert_angles]:
            function(
                molecule,
                tip3p_interchange.topology.molecule(0),
                tip3p_interchange,
            )

        assert len(molecule.settles) == 1
        assert len(molecule.angles) == 0
        assert len(molecule.bonds) == 0



import openmm
import openmm.app
from openff.toolkit import Topology

from openff.interchange.exceptions import UnsupportedImportError

_SUPPORTED_VIRTUAL_SITE_TYPES: set[type] = {
    openmm.ThreeParticleAverageSite,
}


def _check_compatible_inputs(
    system: openmm.System,
    topology: openmm.app.Topology | Topology | None,
):
    """Check that inputs are compatible and supported."""
    for index in range(system.getNumParticles()):
        if system.isVirtualSite(index):
            if type(system.getVirtualSite(index)) not in _SUPPORTED_VIRTUAL_SITE_TYPES:
                raise UnsupportedImportError(
                    f"A particle is a virtual site of type {type(system.getVirtualSite(index))}, "
                    "which is not yet supported.",
                )

    if isinstance(topology, Topology):
        _topology = topology.to_openmm()
    else:
        _topology = topology

    if _topology is not None:
        if system.getNumParticles() != _topology.getNumAtoms():
            raise UnsupportedImportError(
                f"The number of particles in the system ({system.getNumParticles()}) and "
                f"the number of atoms in the topology ({_topology.getNumAtoms()}) do not match.",
            )



import random
import re
import time

import numpy
import pytest
from openff.toolkit import ForceField, Molecule, Quantity, RDKitToolkitWrapper, Topology
from openff.toolkit.typing.engines.smirnoff import (
    ChargeIncrementModelHandler,
    ElectrostaticsHandler,
    LibraryChargeHandler,
    NAGLChargesHandler,
    ToolkitAM1BCCHandler,
    vdWHandler,
)
from openff.toolkit.utils.exceptions import MissingPackageError, SMIRNOFFVersionError
from openff.toolkit.utils.toolkit_registry import ToolkitRegistry, toolkit_registry_manager
from openff.utilities import skip_if_missing
from packaging.version import Version

from openff.interchange import Interchange
from openff.interchange.exceptions import NonIntegralMoleculeChargeError
from openff.interchange.smirnoff._nonbonded import (
    SMIRNOFFElectrostaticsCollection,
    _downconvert_vdw_handler,
    _upconvert_vdw_handler,
)


class TestNonbonded:
    def test_electrostatics_am1_handler(self, methane):
        methane.assign_partial_charges(partial_charge_method="am1bcc")

        # Explicitly store these, since results differ RDKit/AmberTools vs. OpenEye
        reference_charges = [c.m for c in methane.partial_charges]

        parameter_handlers = [
            ElectrostaticsHandler(version=0.3),
            ToolkitAM1BCCHandler(version=0.3),
        ]

        electrostatics_handler = SMIRNOFFElectrostaticsCollection.create(
            parameter_handlers,
            methane.to_topology(),
        )
        numpy.testing.assert_allclose(
            [charge.m_as("e") for charge in electrostatics_handler._get_charges().values()],
            reference_charges,
        )

    def test_electrostatics_library_charges(self, methane):
        library_charge_handler = LibraryChargeHandler(version=0.3)
        library_charge_handler.add_parameter(
            {
                "smirks": "[#6X4:1]-[#1:2]",
                "charge1": Quantity(-0.1, "elementary_charge"),
                "charge2": Quantity(0.025, "elementary_charge"),
            },
        )

        parameter_handlers = [
            ElectrostaticsHandler(version=0.3),
            library_charge_handler,
        ]

        electrostatics_handler = SMIRNOFFElectrostaticsCollection.create(
            parameter_handlers,
            methane.to_topology(),
        )

        numpy.testing.assert_allclose(
            [charge.m_as("e") for charge in electrostatics_handler._get_charges().values()],
            [-0.1, 0.025, 0.025, 0.025, 0.025],
        )

    def test_electrostatics_charge_increments(self, hydrogen_chloride):
        hydrogen_chloride.assign_partial_charges(partial_charge_method="am1-mulliken")

        reference_charges = [c.m for c in hydrogen_chloride.partial_charges]
        reference_charges[0] += 0.1
        reference_charges[1] -= 0.1

        charge_increment_handler = ChargeIncrementModelHandler(version=0.3)
        charge_increment_handler.add_parameter(
            {
                "smirks": "[#17:1]-[#1:2]",
                "charge_increment1": Quantity(0.1, "elementary_charge"),
                "charge_increment2": Quantity(-0.1, "elementary_charge"),
            },
        )

        parameter_handlers = [
            ElectrostaticsHandler(version=0.3),
            charge_increment_handler,
        ]

        electrostatics_handler = SMIRNOFFElectrostaticsCollection.create(
            parameter_handlers,
            hydrogen_chloride.to_topology(),
        )

        # AM1-Mulliken charges are [-0.168,  0.168], increments are [0.1, -0.1],
        # sum is [-0.068,  0.068]
        numpy.testing.assert_allclose(
            [charge.m_as("e") for charge in electrostatics_handler._get_charges().values()],
            reference_charges,
        )

    @pytest.mark.slow
    def test_toolkit_am1bcc_uses_elf10_if_oe_is_available(self, sage_no_nagl, hexane_diol):
        """
        Ensure that the ToolkitAM1BCCHandler assigns ELF10 charges if OpenEye is available.

        Taken from https://github.com/openforcefield/openff-toolkit/pull/1214,
        """
        try:
            hexane_diol.assign_partial_charges(partial_charge_method="am1bccelf10")
            uses_elf10 = True
        except ValueError:
            # This assumes that the ValueError stems from "am1bccelf10" and not other sources; the
            # toolkit should implement a failure mode that does not clash with other `ValueError`s
            hexane_diol.assign_partial_charges(partial_charge_method="am1bcc")
            uses_elf10 = False

        partial_charges = [c.m for c in hexane_diol.partial_charges]

        assigned_charges = [
            v.m
            for v in Interchange.from_smirnoff(sage_no_nagl, [hexane_diol])["Electrostatics"]._get_charges().values()
        ]

        try:
            from openeye import oechem

            openeye_available = oechem.OEChemIsLicensed()
        except ImportError:
            openeye_available = False

        if openeye_available:
            assert uses_elf10
            numpy.testing.assert_allclose(partial_charges, assigned_charges)
        else:
            assert not uses_elf10
            numpy.testing.assert_allclose(partial_charges, assigned_charges)

    def test_nagl_charge_assignment_matches_reference(self, sage, hexane_diol):
        pytest.importorskip("openff.nagl")

        hexane_diol.assign_partial_charges("openff-gnn-am1bcc-0.1.0-rc.3.pt")
        # Leave the ToolkitAM1BCC tag in openff-2.1.0 to ensure that the NAGLCharges handler takes precedence

        interchange = sage.create_interchange(topology=hexane_diol.to_topology())

        assigned_charges_unitless = interchange["Electrostatics"].get_charge_array().m

        expected_charges = hexane_diol.partial_charges
        assert expected_charges is not None
        assert str(expected_charges.units) == "elementary_charge"
        assert not all(charge == 0 for charge in expected_charges.magnitude)
        expected_charges_unitless = [v.m for v in expected_charges]
        numpy.testing.assert_allclose(expected_charges_unitless, assigned_charges_unitless)


@skip_if_missing("openff.nagl")
class TestNAGLChargesErrorHandling:
    """Test NAGLCharges error conditions."""

    def test_nagl_charges_missing_toolkit_error(self, sage, hexane_diol):
        """Test MissingPackageError when NAGL toolkit is not available. This should fail immediately instead of falling
        back to ToolkitAM1BCC, since it doesn't know whether the molecule would have successfully
        had charges assigned by NAGL if it were available."""

        # Mock the toolkit registry to not have NAGL
        # RDKit is needed for SMARTS matching.
        with toolkit_registry_manager(ToolkitRegistry(toolkit_precedence=[RDKitToolkitWrapper])):
            with pytest.raises(MissingPackageError, match="NAGL software isn't present"):
                sage.create_interchange(topology=hexane_diol.to_topology())

            # No error should be raised if using charge_from_molecules
            sage.create_interchange(
                topology=hexane_diol.to_topology(),
                charge_from_molecules=[hexane_diol],
            )

    def test_nagl_charges_invalid_model_file(self, sage_no_nagl, hexane_diol):
        """Test error handling for invalid model file paths. This should fail immediately instead of falling
        back to ToolkitAM1BCC, since it doesn't know whether the molecule would have successfully
        had charges assigned by this model it had been able to find it."""

        sage_no_nagl.get_parameter_handler(
            "NAGLCharges",
            {
                "model_file": "nonexistent_model.pt",
                "version": "0.3",
            },
        )
        with pytest.raises(FileNotFoundError):
            sage_no_nagl.create_interchange(topology=hexane_diol.to_topology())

        sage_no_nagl["NAGLCharges"].model_file = ""
        with pytest.raises(FileNotFoundError):
            sage_no_nagl.create_interchange(topology=hexane_diol.to_topology())

        sage_no_nagl["NAGLCharges"].model_file = None
        with pytest.raises(FileNotFoundError):
            sage_no_nagl.create_interchange(topology=hexane_diol.to_topology())

    def test_nagl_charges_bad_hash(self, sage_no_nagl, hexane_diol, monkeypatch):
        """Test error handling for a bad hash. This should fail immediately instead of falling
        back to ToolkitAM1BCC, since it doesn't know whether the molecule would have successfully
        had charges assigned by this model if the hash comparison hadn't failed."""

        from openff.nagl_models._dynamic_fetch import HashComparisonFailedException

        sage_no_nagl.get_parameter_handler(
            "NAGLCharges",
            {
                "model_file": "openff-gnn-am1bcc-0.1.0-rc.3.pt",
                "model_file_hash": "bad_hash",
                "version": "0.3",
            },
        )
        with pytest.raises(HashComparisonFailedException):
            sage_no_nagl.create_interchange(topology=hexane_diol.to_topology())

    def test_nagl_charges_bad_doi(self, sage_no_nagl, hexane_diol, monkeypatch):
        """Test error handling for a bad DOI. This should fail immediately instead of falling
        back to ToolkitAM1BCC, since it doesn't know whether the molecule would have successfully
        had charges assigned by this model, since it's unfetchable."""

        from openff.nagl_models._dynamic_fetch import UnableToParseDOIException

        sage_no_nagl.get_parameter_handler(
            "NAGLCharges",
            {
                "model_file": "nonexistent_model.pt",
                "digital_object_identifier": "blah.foo/bar",
                "version": "0.3",
            },
        )
        with pytest.raises(UnableToParseDOIException):
            sage_no_nagl.create_interchange(topology=hexane_diol.to_topology())

    # For more information on why this test is skipped, see
    # https://github.com/openforcefield/openff-interchange/pull/1206/commits/f99a10e17ad56235ba1f36ae35f6383a22ed840a#r2248864028
    @pytest.mark.xfail(
        reason="charge assignment handler fallback behavior not yet implemented",
        raises=ValueError,
    )
    def test_nagl_charges_fallback_to_charge_increment_model(self, sage):
        """Test that NAGL falls back to ChargeIncrementModel when molecule contains unsupported elements."""

        # Create a boron-containing molecule with nonzero formal charge
        # BF4- anion - boron is not supported by current NAGL models
        boron_molecule = Molecule.from_smiles("[B-]([F])([F])([F])[F]")

        # Verify formal charges are not all zero
        formal_charges = [atom.formal_charge.m for atom in boron_molecule.atoms]
        assert not all(charge == 0 for charge in formal_charges)

        # Create minimal force field with only the needed handlers
        ff = ForceField()

        # Add Electrostatics handler
        ff.register_parameter_handler(
            ElectrostaticsHandler(version="0.4"),
        )

        # Add NAGLCharges handler
        ff.register_parameter_handler(
            NAGLChargesHandler(
                version="0.3",
                model_file="openff-gnn-am1bcc-0.1.0-rc.3.pt",
            ),
        )

        # Add ChargeIncrementModel handler with formal_charge method and no increments
        charge_increment_handler = ChargeIncrementModelHandler(
            version="0.3",
            partial_charge_method="formal_charge",
        )
        ff.register_parameter_handler(charge_increment_handler)

        # Should succeed despite NAGL not supporting boron
        interchange = ff.create_interchange(topology=boron_molecule.to_topology())

        # Should have assigned charges to all atoms
        assigned_charges = interchange["Electrostatics"].get_charge_array()

        # Assigned charges should match formal charges (fallback to ChargeIncrementModel)
        expected_charges = [atom.formal_charge.m for atom in boron_molecule.atoms]
        numpy.testing.assert_allclose(assigned_charges.m, expected_charges)

        # Net charge should match molecule's total formal charge
        assert abs(sum(assigned_charges.m) - boron_molecule.total_charge.m) < 1e-10

    @pytest.mark.xfail(
        reason="charge assignment handler fallback behavior not yet implemented",
        raises=ValueError,
    )
    def test_nagl_charges_all_handlers_fail_comprehensive_error(self, sage):
        """Test error reporting when all charge assignment methods fail."""

        # Create a uranium compound - not supported by any current charge assignment method
        uranium_molecule = Molecule.from_smiles("[U+4]")

        # Create force field with multiple charge assignment handlers
        ff = ForceField()

        # Add Electrostatics handler
        ff.register_parameter_handler(
            ElectrostaticsHandler(version="0.4"),
        )

        # Add NAGLCharges handler
        ff.register_parameter_handler(
            NAGLChargesHandler(
                version="0.3",
                model_file="openff-gnn-am1bcc-0.1.0-rc.3.pt",
            ),
        )

        # Add ToolkitAM1BCC handler
        ff.register_parameter_handler(
            ToolkitAM1BCCHandler(version="0.3"),
        )

        # Add ChargeIncrementModel handler with gasteiger method
        charge_increment_handler = ChargeIncrementModelHandler(
            version="0.3",
            partial_charge_method="mmff94",
        )
        ff.register_parameter_handler(charge_increment_handler)

        # Should fail with comprehensive error message
        with pytest.raises(RuntimeError) as excinfo:
            ff.create_interchange(topology=uranium_molecule.to_topology())

        error_message = str(excinfo.value)

        # Error should mention that no charges could be assigned
        assert "could not be fully assigned charges" in error_message

        # Error should contain information about each handler's failure
        assert "NAGLCharges" in error_message
        assert "ToolkitAM1BCC" in error_message
        assert "ChargeIncrementModel" in error_message

        # Should mention the exceptions raised by various handlers
        assert "exceptions raised by various handlers" in error_message


@skip_if_missing("openff.nagl")
class TestNAGLChargesPrecedence:
    """Test NAGLCharges precedence in the hierarchy of charge assignment methods."""

    def test_nagl_charges_precedence_over_am1bcc(self, sage, hexane_diol):
        """Test that NAGLCharges takes precedence over ToolkitAM1BCC."""

        sage.get_parameter_handler("ToolkitAM1BCC", {"version": "0.3"})
        # Get reference charges from NAGL
        hexane_diol.assign_partial_charges("openff-gnn-am1bcc-0.1.0-rc.3.pt")
        nagl_charges = [c.m for c in hexane_diol.partial_charges]

        # Get reference charges from AM1BCC
        hexane_diol.assign_partial_charges("am1bcc")
        am1bcc_charges = [c.m for c in hexane_diol.partial_charges]

        # Ensure they're different
        assert not numpy.allclose(nagl_charges, am1bcc_charges)

        interchange = sage.create_interchange(topology=hexane_diol.to_topology())
        assigned_charges = interchange["Electrostatics"].get_charge_array()

        # Should match NAGL charges, not AM1BCC
        numpy.testing.assert_allclose(assigned_charges, nagl_charges)

    def test_library_charges_precedence_over_nagl(self, sage, methane):
        """Test that LibraryCharges takes precedence over NAGLCharges."""

        sage["LibraryCharges"].add_parameter(
            {
                "smirks": "[#6X4:1]-[#1:2]",
                "charge1": Quantity(-0.2, "elementary_charge"),
                "charge2": Quantity(0.05, "elementary_charge"),
            },
        )

        interchange = sage.create_interchange(topology=methane.to_topology())
        assigned_charges = interchange["Electrostatics"].get_charge_array()

        # Should match library charges
        expected_charges = [-0.2, 0.05, 0.05, 0.05, 0.05]
        numpy.testing.assert_allclose(assigned_charges, expected_charges)

    def test_nagl_charges_precedence_over_charge_increments(self, sage, hexane_diol):
        """Test that NAGLCharges takes precedence over ChargeIncrementModel as base charges."""

        # Get reference charges from NAGL
        hexane_diol.assign_partial_charges("openff-gnn-am1bcc-0.1.0-rc.3.pt")
        nagl_charges = [c.m for c in hexane_diol.partial_charges]

        # Add ChargeIncrementModel handler (should provide base charges, not increments)
        increment_handler = ChargeIncrementModelHandler(
            version=0.3,
            partial_charge_method="formal_charge",
        )
        sage.register_parameter_handler(increment_handler)

        interchange = sage.create_interchange(topology=hexane_diol.to_topology())
        assigned_charges = interchange["Electrostatics"].get_charge_array()

        # Should match NAGL charges, not formal charges
        numpy.testing.assert_allclose(assigned_charges, nagl_charges)


@skip_if_missing("openff.nagl")
class TestNAGLChargesIntegration:
    """Test NAGLCharges integration with other handlers."""

    def test_nagl_charges_multi_molecule_topology(self, sage):
        """Test NAGLCharges with multiple molecules in topology."""

        methane = Molecule.from_smiles("C")
        ethane = Molecule.from_smiles("CC")

        topology = Topology.from_molecules([methane, ethane])

        interchange = sage.create_interchange(topology=topology)
        assigned_charges = interchange["Electrostatics"].get_charge_array()

        # Each molecule should have approximately zero net charge
        methane_charge_sum = sum(assigned_charges[: methane.n_atoms])
        ethane_charge_sum = sum(assigned_charges[methane.n_atoms :])

        assert abs(methane_charge_sum) < Quantity(1e-10, "elementary_charge")
        assert abs(ethane_charge_sum) < Quantity(1e-10, "elementary_charge")

    def test_nagl_charges_with_virtual_sites(self, sage_with_bond_charge):
        """Test NAGLCharges compatibility with virtual sites."""

        # Create a molecule that would have virtual sites
        molecule = Molecule.from_smiles("[Cl]CCO")

        # Should not raise an error
        interchange = sage_with_bond_charge.create_interchange(
            topology=molecule.to_topology(),
        )

        # Should have charges for real atoms
        assigned_charges = interchange["Electrostatics"]._get_charges()
        assert len(assigned_charges.values()) - 1 == molecule.n_atoms

        # Net charge should be approximately zero
        all_particle_charge_sum = sum(assigned_charges.values())
        assert abs(all_particle_charge_sum) < Quantity(1e-10, "elementary_charge")
        # Charge without the vsite should be nonzero
        atom_charge_sum = sum([charge for tk, charge in assigned_charges.items() if tk.atom_indices is not None])
        assert abs(atom_charge_sum - Quantity(0.123, "elementary_charge")) < Quantity(1e-10, "elementary_charge")

    def test_nagl_charges_force_field_creation_complete(self, hexane_diol):
        """Test complete interchange creation with NAGLCharges."""

        ff = ForceField("openff-2.1.0.offxml")
        ff.get_parameter_handler(
            "NAGLCharges",
            {
                "model_file": "openff-gnn-am1bcc-0.1.0-rc.3.pt",
                "version": "0.3",
            },
        )

        # Should create complete interchange without errors
        interchange = ff.create_interchange(topology=hexane_diol.to_topology())

        # Should have all expected collections
        expected_collections = ["Bonds", "Angles", "ProperTorsions", "ImproperTorsions", "vdW", "Electrostatics"]
        for collection_name in expected_collections:
            assert collection_name in interchange.collections

        # Electrostatics should have charges
        charges = interchange["Electrostatics"].get_charge_array()
        assert len(charges) == hexane_diol.n_atoms

        # Net charge should be approximately zero
        total_charge = sum(charge.m for charge in charges)
        assert abs(total_charge) < 1e-10

    def test_nagl_charges_identical_molecules_same_charges(self):
        """Test that identical molecules get identical charges from NAGLCharges."""

        # Create topology with two identical molecules
        molecule1 = Molecule.from_smiles("CCO")
        molecule2 = Molecule.from_smiles("CCO")
        topology = Topology.from_molecules([molecule1, molecule2])

        ff = ForceField("openff-2.1.0.offxml")
        ff.get_parameter_handler(
            "NAGLCharges",
            {
                "model_file": "openff-gnn-am1bcc-0.1.0-rc.3.pt",
                "version": "0.3",
            },
        )

        interchange = ff.create_interchange(topology=topology)
        assigned_charges = interchange["Electrostatics"].get_charge_array()

        # First molecule charges
        mol1_charges = assigned_charges[: molecule1.n_atoms]
        # Second molecule charges
        mol2_charges = assigned_charges[molecule1.n_atoms :]

        # Should be identical
        numpy.testing.assert_allclose(mol1_charges, mol2_charges)

    def test_nagl_charges_with_charge_from_molecules(self, sage, hexane_diol):
        """Test that charge_from_molecules takes precedence over NAGLCharges."""

        # Assign preset charges using a different method
        hexane_diol.assign_partial_charges("gasteiger")
        preset_charges = [c.m for c in hexane_diol.partial_charges]

        # Create interchange with charge_from_molecules - should use preset charges
        interchange = sage.create_interchange(
            topology=hexane_diol.to_topology(),
            charge_from_molecules=[hexane_diol],
        )

        assigned_charges = interchange["Electrostatics"].get_charge_array()

        # Should match preset charges, not NAGL charges
        numpy.testing.assert_allclose(assigned_charges.m, preset_charges)

        # Verify NAGL would give different charges
        hexane_diol_copy = Molecule.from_smiles(hexane_diol.to_smiles())
        hexane_diol_copy.assign_partial_charges("openff-gnn-am1bcc-0.1.0-rc.3.pt")
        nagl_charges = [c.m for c in hexane_diol_copy.partial_charges]

        # Preset and NAGL charges should be different
        assert not numpy.allclose(preset_charges, nagl_charges, atol=1e-3)

    def test_nagl_charges_with_mixed_charge_sources(self, sage):
        """Test NAGLCharges with some molecules having preset charges and others not."""

        # Create molecules
        ethanol = Molecule.from_smiles("CCO")
        methanol = Molecule.from_smiles("CO")

        # Assign preset charges to only one molecule
        ethanol.assign_partial_charges("gasteiger")
        preset_ethanol_charges = [c.m for c in ethanol.partial_charges]

        topology = Topology.from_molecules([ethanol, methanol])

        # Create interchange with preset charges for ethanol only
        interchange = sage.create_interchange(
            topology=topology,
            charge_from_molecules=[ethanol],
        )

        assigned_charges = interchange["Electrostatics"].get_charge_array()

        # First molecule (ethanol) should match preset charges
        ethanol_charges = assigned_charges[: ethanol.n_atoms]
        numpy.testing.assert_allclose(ethanol_charges.m, preset_ethanol_charges)

        # Second molecule (methanol) should get NAGL charges
        methanol_charges = assigned_charges[ethanol.n_atoms :]

        # Get reference NAGL charges for methanol
        methanol_copy = Molecule.from_smiles("CO")
        methanol_copy.assign_partial_charges("openff-gnn-am1bcc-0.1.0-rc.3.pt")
        nagl_methanol_charges = [c.m for c in methanol_copy.partial_charges]

        numpy.testing.assert_allclose(methanol_charges.m, nagl_methanol_charges)

    @pytest.mark.slow
    def test_nagl_charges_large_molecule_performance(self, sage):
        """Test that NAGL charge assignment completes in reasonable time for large molecules."""

        # Create a very large molecule
        large_molecule = Molecule.from_smiles("C" * 200)  # 200-carbon alkane chain

        start_time = time.time()

        # Should complete without error
        interchange = sage.create_interchange(topology=large_molecule.to_topology())

        end_time = time.time()
        execution_time = end_time - start_time

        # Should complete within reasonable time (less than 30 seconds)
        assert execution_time < 60.0, f"NAGL charge assignment took {execution_time:.2f}s, which is too long"

        # Net charge should be approximately zero
        charges = interchange["Electrostatics"].get_charge_array()
        total_charge = sum(charges.m)
        assert abs(total_charge) < 1e-10

    @pytest.mark.slow
    def test_nagl_charges_multiple_large_molecules_performance(self, sage):
        """Test performance with multiple large molecules in topology."""
        # Create multiple copies of medium-sized molecules
        base_molecules = [
            Molecule.from_smiles("C" * 20),  # 20-carbon chain
            Molecule.from_smiles("C" * 25),  # 25-carbon chain
            Molecule.from_smiles("C" * 30),  # 30-carbon chain
        ]

        # Create 20 copies of each
        molecules = []
        for _ in range(20):
            for base_mol in base_molecules:
                molecules.append(base_mol)

        topology = Topology.from_molecules(molecules)

        start_time = time.time()

        # Should complete without error
        interchange = sage.create_interchange(topology=topology)

        end_time = time.time()
        execution_time = end_time - start_time

        # Should complete within reasonable time
        assert execution_time < 60.0, f"Multi-molecule NAGL assignment took {execution_time:.2f}s, which is too long"

        # Each molecule should have approximately zero net charge
        charges = interchange["Electrostatics"].get_charge_array()
        start_idx = 0
        for molecule in molecules:
            mol_charges = charges[start_idx : start_idx + molecule.n_atoms]
            mol_total_charge = sum(mol_charges.m)
            assert abs(mol_total_charge) < 1e-10
            start_idx += molecule.n_atoms

    @pytest.mark.skip(
        reason="Turn on if toolkit ever allows non-standard scale12/13/15",
    )
    def test_nonstandard_scale_factors(
        sage,
        methane,
    ):
        # SMIRNOFFSpecError: Current OFF toolkit is unable to handle scale12 values other than 0.0.
        # SMIRNOFFSpecError: Current OFF toolkit is unable to handle scale13 values other than 0.0.
        # SMIRNOFFSpecError: Current OFF toolkit is unable to handle scale15 values other than 1.0.

        factors = {index: random.random() for index in range(2, 6)}

        for handler in ("vdW", "Electrostatics"):
            for index, factor in factors.items():
                setattr(sage[handler], f"scale1{index}", factor)

        interchange = sage.create_interchange(methane.to_topology())

        for collection in ("vdW", "Electrostatics"):
            for index, factor in factors.items():
                assert getattr(interchange[collection], f"scale1{index}") == factor


class TestvdWUpDownConversion:
    def test_upconversion(self):
        handler = vdWHandler(version=0.3, method="cutoff")

        if handler.version == Version("0.4"):
            pytest.skip("Don't test upconversion if the toolkit already did it")

        try:
            _upconvert_vdw_handler(handler)
        except SMIRNOFFVersionError:
            pytest.skip("The installed version of the toolkit does not support 0.4")

        assert handler.version == Version("0.4")
        assert handler.periodic_method == "cutoff"
        assert handler.nonperiodic_method == "no-cutoff"

    def test_downconversion(self):
        try:
            handler = vdWHandler(version=0.4)
        except SMIRNOFFVersionError:
            pytest.skip("The installed version of the toolkit does not support 0.4")

        _downconvert_vdw_handler(handler)

        assert handler.version == Version("0.3")
        assert handler.method == "cutoff"

        # Update when https://github.com/openforcefield/openff-toolkit/issues/1680 is resolved
        try:
            assert not hasattr(handler, "nonperiodic_method")
            assert not hasattr(handler, "periodic_method")
        except AssertionError:
            assert not hasattr(handler, "__delete__")
            pytest.skip("ParameterAttribute.__delete__ not implemented")


class TestElectrostatics:
    def test_caching_detects_atom_ordering(self, sage):
        def get_charges_from_interchange(
            molecule: Molecule,
        ) -> dict[int, Quantity]:
            return {
                key.atom_indices[0]: val
                for key, val in sage.create_interchange(molecule.to_topology())["Electrostatics"]
                ._get_charges()
                .items()
            }

        def compare_charges(
            molecule: Molecule,
            interchange_charges: dict[int, Quantity],
        ):
            for index, molecule_charge in enumerate(molecule.partial_charges):
                assert interchange_charges[index] == molecule_charge

        original = Molecule.from_mapped_smiles("[H:1]-[C:2]#[N:3]")
        reordered = Molecule.from_mapped_smiles("[H:3]-[C:2]#[N:1]")

        for molecule in [original, reordered]:
            molecule.assign_partial_charges("openff-gnn-am1bcc-1.0.0.pt")

        compare_charges(original, get_charges_from_interchange(original))
        compare_charges(reordered, get_charges_from_interchange(reordered))

    def test_get_charge_array(self, sage):
        ammonia = Molecule.from_smiles("N")
        ammonia.partial_charges = Quantity(numpy.array([-3, 1, 1, 1]) / 3, "elementary_charge")

        # topology is water | ethanol         | ammonia
        #              O H H C C O H H H H H H N H H H
        topology = Topology.from_molecules(
            [
                Molecule.from_mapped_smiles("[H:2][O:1][H:3]"),
                Molecule.from_mapped_smiles("[H:4][C:1]([H:5])([H:6])[C:2]([H:7])([H:8])[O:3][H:9]"),
                Molecule.from_mapped_smiles("[H:2][N:1]([H:3])[H:4]"),
            ],
        )

        charges = sage.create_interchange(topology, charge_from_molecules=[ammonia])[
            "Electrostatics"
        ].get_charge_array()

        assert isinstance(charges, Quantity)
        assert not isinstance(charges, numpy.ndarray)
        assert isinstance(charges.m, numpy.ndarray)

        # TIP3P library charges
        assert numpy.allclose(
            charges[:3].m,
            [-0.834, 0.417, 0.417],
        )

        # these are from openff-gnn-am1bcc-1.0.0.pt
        assert numpy.allclose(
            charges[3:12].m,
            [
                -0.09629,
                0.13245,
                -0.60293001,
                0.04465,
                0.04465,
                0.04465,
                0.01728,
                0.01728,
                0.39826,
            ],
        )

        # these are the hard-coded 1, and -1/3
        assert numpy.allclose(
            charges[-4:].m,
            ammonia.partial_charges.m,
        )


def test_get_charge_array_fails_if_virtual_sites_present(water, tip4p):
    with pytest.raises(
        NotImplementedError,
        match="Not yet implemented with virtual sites",
    ):
        tip4p.create_interchange(water.to_topology())["Electrostatics"].get_charge_array(include_virtual_sites=True)

    with pytest.raises(
        NotImplementedError,
        match="Not yet implemented when virtual sites are present",
    ):
        tip4p.create_interchange(water.to_topology())["Electrostatics"].get_charge_array(include_virtual_sites=False)


def test_nonintegral_molecule_charge_error(sage, water):
    funky_charges = Quantity([0, 0, -5.5], "elementary_charge")

    water.partial_charges = funky_charges

    with pytest.raises(
        NonIntegralMoleculeChargeError,
        match=re.escape("net charge of -5.5 compared to a total formal charge of 0.0"),
    ):
        sage.create_interchange(water.to_topology(), charge_from_molecules=[water])


class TestSMIRNOFFChargeIncrements:
    @pytest.fixture
    def hydrogen_cyanide_charge_increments(self):
        handler = ChargeIncrementModelHandler(
            version=0.4,
            partial_charge_method="formal_charge",
        )
        handler.add_parameter(
            {
                "smirks": "[H:1][C:2]",
                "charge_increment1": Quantity(-0.111, "elementary_charge"),
                "charge_increment2": Quantity(0.111, "elementary_charge"),
            },
        )
        handler.add_parameter(
            {
                "smirks": "[C:1]#[N:2]",
                "charge_increment1": Quantity(0.5, "elementary_charge"),
                "charge_increment2": Quantity(-0.5, "elementary_charge"),
            },
        )

        return handler

    def test_no_charge_increments_applied(self, sage, hexane_diol):
        gastiger_charges = [c.m for c in hexane_diol.partial_charges]
        sage.deregister_parameter_handler("NAGLCharges")

        no_increments = ChargeIncrementModelHandler(
            version=0.3,
            partial_charge_method="gasteiger",
        )
        sage.register_parameter_handler(no_increments)

        assert len(sage["ChargeIncrementModel"].parameters) == 0

        out = Interchange.from_smirnoff(sage, [hexane_diol])
        assert numpy.allclose(
            numpy.asarray([v.m for v in out["Electrostatics"]._get_charges().values()]),
            gastiger_charges,
        )

    def test_overlapping_increments(self, sage, methane):
        """Test that separate charge increments can be properly applied to the same atom."""
        sage.deregister_parameter_handler("NAGLCharges")
        charge_handler = ChargeIncrementModelHandler(
            version=0.3,
            partial_charge_method="formal_charge",
        )
        charge_handler.add_parameter(
            {
                "smirks": "[C:1][H:2]",
                "charge_increment1": Quantity(0.111, "elementary_charge"),
                "charge_increment2": Quantity(-0.111, "elementary_charge"),
            },
        )
        sage.register_parameter_handler(charge_handler)
        assert 0.0 == pytest.approx(
            sum(v.m for v in Interchange.from_smirnoff(sage, [methane])["Electrostatics"]._get_charges().values()),
        )

    def test_charge_increment_forwawrd_reverse_molecule(
        self,
        sage,
        hydrogen_cyanide,
        hydrogen_cyanide_reversed,
        hydrogen_cyanide_charge_increments,
    ):
        sage.deregister_parameter_handler("NAGLCharges")
        sage.register_parameter_handler(hydrogen_cyanide_charge_increments)

        topology = Topology.from_molecules(
            [hydrogen_cyanide, hydrogen_cyanide_reversed],
        )

        out = sage.create_interchange(topology)

        expected_charges = [-0.111, 0.611, -0.5, -0.5, 0.611, -0.111]

        # TODO: Fix get_charges to return the atoms in order
        found_charges = [0.0] * topology.n_atoms
        for key, val in out["Electrostatics"]._get_charges().items():
            found_charges[key.atom_indices[0]] = val.m

        assert numpy.allclose(expected_charges, found_charges)



from typing import Literal

import openmm
from openff.toolkit import Quantity, Topology
from pydantic import Field

from openff.interchange.components.potentials import Collection, Potential
from openff.interchange.models import PotentialKey, VirtualSiteKey


class _VirtualSiteCollection(Collection):
    """
    A handler which stores the information necessary to construct non-SMIRNOFF virtual sites.
    """

    key_map: dict[VirtualSiteKey, PotentialKey] = Field(
        dict(),
        description="A mapping between VirtualSiteKey objects and PotentialKey objects.",
    )  # type: ignore[assignment]

    type: Literal["VirtualSites"] = "VirtualSites"
    expression: Literal[""] = ""
    virtual_site_key_topology_index_map: dict[VirtualSiteKey, int] = Field(
        dict(),
        description="A mapping between VirtualSiteKey objects (stored analogously to TopologyKey objects"
        "in other handlers) and topology indices describing the associated virtual site",
    )

    exclusion_policy: Literal["parents"] = "parents"


def _convert_virtual_sites(
    system: openmm.System,
    topology: Topology,
    particle_map: dict,
) -> _VirtualSiteCollection | None:
    if particle_map is None:
        return None

    collection = _VirtualSiteCollection()

    for particle_index in range(system.getNumParticles()):
        if system.getParticleMass(particle_index)._value != 0.0:
            continue

        virtual_site = system.getVirtualSite(particle_index)

        openmm_particle_indices: tuple[int, int, int] = (
            virtual_site.getParticle(0),
            virtual_site.getParticle(1),
            virtual_site.getParticle(2),
        )

        # get ImportedVirtualSiteKey from topology._molecule_virtual_site_map
        for _, list_of_virtual_sites in topology._molecule_virtual_site_map.items():
            for virtual_site_key in list_of_virtual_sites:
                if virtual_site_key.orientation_atom_indices == tuple(
                    particle_map[i] for i in openmm_particle_indices
                ):
                    # TODO: Not the cleanest way to de-duplicate virtual site parameters,
                    #       but other info (i.e. atom name) is insufficient
                    id = type(virtual_site).__name__ + "-".join(
                        map(
                            str,
                            [
                                virtual_site.getWeight(0),
                                virtual_site.getWeight(1),
                                virtual_site.getWeight(2),
                            ],
                        ),
                    )

                    potential_key = PotentialKey(id=id, virtual_site_type="ThreeParticleSite")

                    collection.key_map[virtual_site_key] = potential_key
                    # TODO: This is where geometry-specific information is stored in SMIRNOFF virtual sites,
                    #       but in this case that's not needed
                    collection.potentials[potential_key] = Potential(
                        parameters={
                            "weights": Quantity([virtual_site.getWeight(i) for i in range(3)], "dimensionless"),
                        },
                    )

    return collection if len(collection.key_map) > 0 else None



"""Helpers for serialization/Pydantic things."""

from typing import Annotated

from openff.toolkit import Topology
from pydantic import (
    PlainSerializer,
    SerializerFunctionWrapHandler,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    WrapSerializer,
    WrapValidator,
)

from openff.interchange.common._topology import InterchangeTopology


def _topology_custom_before_validator(
    topology: str | Topology | InterchangeTopology,
    handler: ValidatorFunctionWrapHandler,
    info: ValidationInfo,
) -> InterchangeTopology:
    if info.mode == "json":
        # Making a new one so no need to deepcopy
        return InterchangeTopology.from_json(topology)

    assert info.mode == "python"
    if isinstance(topology, Topology | InterchangeTopology):
        return InterchangeTopology(topology)
    elif isinstance(topology, str):
        return InterchangeTopology.from_json(topology)
    else:
        raise Exception(f"Failed to convert topology of type {type(topology)}")


def _topology_json_serializer(
    topology: InterchangeTopology,
    nxt: SerializerFunctionWrapHandler,
) -> str:
    return topology.to_json()


def _topology_dict_serializer(topology: InterchangeTopology) -> dict:
    return topology.to_dict()


_AnnotatedTopology = Annotated[
    InterchangeTopology,
    WrapValidator(_topology_custom_before_validator),
    PlainSerializer(_topology_dict_serializer, return_type=dict),
    WrapSerializer(_topology_json_serializer, when_used="json"),
]



from openff.toolkit import Topology

from openff.interchange.exceptions import NoPositionsError


class InterchangeTopology(Topology):
    """A subclass of the OpenFF Topology storing no positions."""

    def get_positions(self, *args, **kwargs):
        raise NoPositionsError("Interchange.topology does not store positions. Use Interchange.positions instead.")

    def set_positions(self, *args, **kwargs):
        raise NoPositionsError("Interchange.topology does not store positions. Use Interchange.positions instead.")

    def clear_positions(self, *args, **kwargs):
        raise NoPositionsError("Interchange.topology does not store positions. Use Interchange.positions instead.")



import numpy
import pytest
from openff.toolkit import ForceField, Molecule, Quantity
from openff.utilities import get_data_file_path, has_package, skip_if_missing

from openff.interchange import Interchange
from openff.interchange._tests.unit_tests.smirnoff.test_valence import (
    TestBondOrderInterpolation,
)
from openff.interchange.drivers.openmm import (
    _get_openmm_energies,
    _process,
    get_openmm_energies,
)

if has_package("openmm"):
    import openmm
    import openmm.unit


@skip_if_missing("openmm")
class TestBondOrderInterpolationEnergies(TestBondOrderInterpolation):
    @pytest.mark.slow
    def test_basic_bond_order_interpolation_energies(self, xml_ff_bo_bonds):
        forcefield = ForceField(
            get_data_file_path(
                "test_forcefields/test_forcefield.offxml",
                "openff.toolkit",
            ),
            xml_ff_bo_bonds,
        )

        mol = Molecule.from_file(
            get_data_file_path("molecules/CID20742535_anion.sdf", "openff.toolkit"),
        )
        mol.generate_conformers(n_conformers=1)
        top = mol.to_topology()

        out = Interchange.from_smirnoff(forcefield, top)
        out.box = Quantity([4, 4, 4], "nanometer")
        out.positions = mol.conformers[0]

        interchange_bond_energy = get_openmm_energies(
            out,
            combine_nonbonded_forces=True,
        ).energies["Bond"]

        system = forcefield.create_openmm_system(top)
        toolkit_bond_energy = _process(
            _get_openmm_energies(
                system=system,
                box_vectors=[[4, 0, 0], [0, 4, 0], [0, 0, 4]] * openmm.unit.nanometer,
                positions=mol.conformers[0].to_openmm(),
                round_positions=None,
                platform="Reference",
            ),
            system=system,
            combine_nonbonded_forces=True,
            detailed=False,
        ).energies["Bond"]

        assert abs(interchange_bond_energy - toolkit_bond_energy).m < 1e-2

        new = out.to_openmm(combine_nonbonded_forces=True)
        ref = forcefield.create_openmm_system(top)

        new_k = []
        new_length = []
        for force in new.getForces():
            if type(force) is openmm.HarmonicBondForce:
                for i in range(force.getNumBonds()):
                    new_k.append(force.getBondParameters(i)[3]._value)
                    new_length.append(force.getBondParameters(i)[2]._value)

        ref_k = []
        ref_length = []
        for force in ref.getForces():
            if type(force) is openmm.HarmonicBondForce:
                for i in range(force.getNumBonds()):
                    ref_k.append(force.getBondParameters(i)[3]._value)
                    ref_length.append(force.getBondParameters(i)[2]._value)

        numpy.testing.assert_allclose(ref_k, new_k, rtol=3e-5)



import pytest
from openff.toolkit import Quantity

from openff.interchange._tests import needs_gmx
from openff.interchange._tests._gromacs import gmx_monolithic_vs_itp, gmx_roundtrip


@needs_gmx
@pytest.mark.xfail(reason="Test is very flaky in CI")
def test_ligand_vacuum(caffeine, sage_unconstrained, monkeypatch):
    monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "1")

    topology = caffeine.to_topology()
    topology.box_vectors = Quantity([4, 4, 4], "nanometer")

    gmx_roundtrip(sage_unconstrained.create_interchange(topology))
    gmx_monolithic_vs_itp(sage_unconstrained.create_interchange(topology))


@needs_gmx
def test_water_dimer(water_dimer, tip3p, monkeypatch):
    monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "1")

    gmx_roundtrip(tip3p.create_interchange(water_dimer))
    gmx_monolithic_vs_itp(tip3p.create_interchange(water_dimer))


@pytest.mark.xfail(reason="Test is very flaky in CI")
@needs_gmx
def test_alanine_dipeptide(alanine_dipeptide, ff14sb, monkeypatch):
    monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "1")

    gmx_roundtrip(ff14sb.create_interchange(alanine_dipeptide))
    gmx_monolithic_vs_itp(ff14sb.create_interchange(alanine_dipeptide))
